/**
 * \file bound.hpp
 *
 * \brief Definition of a class for  representing the type of a linear
 * constraint (i.e., equality or  inequality) and its right-hand side:
 * a real number.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Matem&aacute;tica, \n
 * mfsiqueira at mat (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date March 2016
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef __bound__
#define __bound__


/** 
 * \defgroup BC2NameSpace Namespace bc2.
 * @{
 */

/**
 * \namespace bc2
 *
 * \brief The namespace bc2 contains the definition and implementation
 * of a  set of  classes for  computing a C1  spline cubic  curve that
 * passes through  a given planar  channel delimited by  two polygonal
 * chains.
 *
 */

namespace bc2 {

  /**
   * \class Bound
   *
   * \brief  This class  represents the  type of  a constraint  (i.e.,
   * equality or inequality)  and the value of its  right-hand side: a
   * real.
   *
   */
  
  class Bound {
  public:

    // -----------------------------------------------------------------
    //
    // Type definitions
    //
    // -----------------------------------------------------------------

    /**
     * \typedef CONSTRAINTYPE
     *
     * \brief Defines a type for the type of a constraint.
     *
     */
    typedef enum { EQT , LTE , GTE } CONSTRAINTYPE ;

    
  protected:

    // ---------------------------------------------------------------
    //
    // Private data
    //
    // ---------------------------------------------------------------

    CONSTRAINTYPE _ctype ; ///< The type of the constraint associated with this bound.
    double _value ;        ///< The bound value.
    unsigned _row ;        ///< The identifier of the constraint associated with this bound.
    

  public:
  
    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------
    
    /**
     * \fn Bound()
     *
     * \brief Creates an instance of this class.
     *
     */
    Bound()
    :
      _ctype( EQT ) ,
      _value( 0 ) ,
      _row( 0 )
    {
    }

    
    /**
     * \fn Bound( CONSTRAINTYPE type , double value , unsigned row ) 
     *
     * \brief Creates an instance of this class.
     *
     * \param type  The type  of the  constraint associated  with this
     * bound.
     * \param value The value of the bound.
     * \param  row The  identifier of  the constraint  associated with
     * this bound.
     *
     */
    Bound( CONSTRAINTYPE type , double value , unsigned row ) 
    :
      _ctype( type ) ,
      _value( value ) ,
      _row( row )
    {
    }


    /**
     * \fn Bound( const Bound& b )
     *
     * \brief Creates an instance of  this class from another instance
     * of this class.
     *
     * \param b An instance of this class.
     *
     */
    Bound( const Bound& b )
    :
      _ctype( b._ctype ) ,
      _value( b._value ) ,
      _row( b._row )
    {
    }
    
    /**
     * \fn ~Bound( )
     *
     * \brief Releases the memory held by an instance of this class.
     *
     */
    ~Bound( )
    {
    }

    
    /**
     * \fn CONSTRAINTYPE get_type() const 
     *
     * \brief Returns the type of  the constraint associated with this
     * bound.
     *
     *
     * \return The type of the constraint associated with this bound.
     *
     */
    CONSTRAINTYPE get_type() const
    {
      return _ctype ;
    }


    /**
     * \fn double get_value() const 
     *
     * \brief Returns the value of this bound.
     *
     * \return The value of this bound.
     *
     */
    double get_value() const
    {
      return _value ;
    }


    /**
     * \fn unsigned get_row() const 
     *
     * \brief Returns the identifier of the constraint associated with
     * this bound. This identifier corresponds  to the number of a row
     * in the constraint coefficient matrix of a linear program.
     *
     *
     * \return The  identifier of the constraint  associated with this
     * bound.
     *
     */
    unsigned get_row() const
    {
      return _row ;
    }

  } ;

}

/** @} */ //end of group class.

#endif   // __bound__
