/** 
 * \file buildcurve2d.cpp
 *
 * \brief Implementation of a class for threading a C1 spline curve of
 * degree d  through a planar channel  defined by a pair  of polygonal
 * chains.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Matem&aacute;tica, \n
 * mfsiqueira at mat (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date March 2016
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "buildcurve2d.hpp"       // BuildCurve2D

#include "exceptionobject.hpp"    // ExceptionObject

extern "C" {
#include "glpk.h"                 // For all GLPK functions
}

#include <cmath>                  // sqrt, fabs, ceil, floor
#include <cassert>                // assert
#include <sstream>                // std::stringstream
#include <iostream>               // std::cerr, std::endl
#include <vector>                 // std::vector
#include <cfloat>                 // DBL_MAX
#include <cstdlib>                // exit, EXIT_FAILURE


/** 
 * \defgroup BC2NameSpace Namespace bc2.
 * @{
 */

/**
 * \namespace bc2
 *
 * \brief The namespace bc2 contains the definition and implementation
 * of a  set of  classes for  computing a C1  spline cubic  curve that
 * passes through  a given planar  channel delimited by  two polygonal
 * chains.
 *
 */

namespace bc2 {


  /**
   * \fn BuildCurve2D::BuildCurve2D( unsigned np , unsigned nc , bool closed , double* lx , double* ly , double* ux , double* uy , TabulatedFunction* tf ) throw( ExceptionObject )
   *
   * \brief Creates an instance of this class.
   *
   * \param np The number of polynomial pieces of the spline curve.
   * \param  nc The  number of  c-segments  in each  c-piece of  the
   * channel.
   * \param closed A flag to indicate whether the channel is closed.
   * \param lx A  pointer to an array with the  x-coordinates of the
   * lower polygonal chain of the channel.
   * \param ly A  pointer to an array with the  y-coordinates of the
   * lower polygonal chain of the channel.
   * \param ux A  pointer to an array with the  x-coordinates of the
   * upper polygonal chain of the channel.
   * \param uy A  pointer to an array with the  y-coordinates of the
   * upper polygonal chain of the channel.
   * \param tf A pointer to the lower and upper \f$a\f$ functions.
   *
   */
  BuildCurve2D::BuildCurve2D(
                             unsigned np ,
                             unsigned nc ,
                             bool closed ,
                             double* lx ,
                             double* ly ,
                             double* ux ,
                             double* uy ,
                             TabulatedFunction* tf
                            )
    throw( ExceptionObject )
  :
    _np( np ) ,
    _nc( nc )
  {
    if ( tf == 0 ) {
      std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
      ss << "Pointer to function A() is null" ;
      throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
    }

    if ( tf->degree() < 2 ) {
      std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
      ss << "The degree of the spline must be at least 2" ;
      throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
    }

    _closed = closed ;
    _dg = tf->degree() ;

    unsigned nn = ( _closed ) ? ( np * nc ) : ( ( np * nc ) + 1 ) ;

    _lxcoords.resize( nn ) ;
    _lycoords.resize( nn ) ;
    _uxcoords.resize( nn ) ;
    _uycoords.resize( nn ) ;

    for ( unsigned i = 0 ; i < nn ; i++ ) {
      _lxcoords[ i ] = lx[ i ] ;
      _lycoords[ i ] = ly[ i ] ;
      _uxcoords[ i ] = ux[ i ] ;
      _uycoords[ i ] = uy[ i ] ;
    }
    
    _tf = tf ;

    _ofvalue = DBL_MAX ;

    return ;
  }


  /**
   * \fn BuildCurve2D::BuildCurve2D( const BuildCurve2D& b )
   *
   * \brief Clones an instance of this class.
   *
   * \param b A reference to another instance of this class.
   *
   */
  BuildCurve2D::BuildCurve2D( const BuildCurve2D& b )
  :
    _np( b._np ) ,
    _nc( b._nc ) ,
    _dg( b._dg ) ,
    _closed( b._closed ) ,
    _lxcoords( b._lxcoords ) ,
    _lycoords( b._lycoords ) ,
    _uxcoords( b._uxcoords ) ,
    _uycoords( b._uycoords ) ,
    _tf( b._tf ) ,
    _coefficients( b._coefficients ) ,
    _bounds( b._bounds ) ,
    _ctrlpts( b._ctrlpts ) ,
    _secdiff( b._secdiff ) ,
    _ofvalue( b._ofvalue )
  {
  }
  
    
  /**
   * \fn BuildCurve2D::~BuildCurve2D()
   *
   * \brief Releases the memory held by an instance of this class.
   *
   */
  BuildCurve2D::~BuildCurve2D()
  {
  }


  /**
   * \fn bool BuildCurve2D::build( int& error )
   *
   * \brief Solves the channel problem by solving a linear program.
   *
   * \param error Code returned by the LP solver whenever a solution
   * could not be  found. If a solution is found,  this parameter is
   * ignored.
   *
   * \return The logic  value true if the LP solver  is able to find
   * an  optimal solution  for the  channel problem;  otherwise, the
   * logic value false is returned.
   *
   */
  bool
  BuildCurve2D::build( int& error )
  {
    sizetype rows = ( sizetype )
                (
                    ( _np * 6 * ( _dg - 1 ) )                                     // min-max
                  + ( ( 2 * _np ) + ( ( !_closed ) ? 2 : 0 ) )                    // correspondence
                  + ( 2 * ( _np - 1 ) + ( ( _closed ) ? 2 : 0 ) )                 // c0
                  + ( 2 * ( _np - 1 ) + ( ( _closed ) ? 2 : 0 ) )                 // c1
                  + ( 4 * ( _np ) * ( _dg + 1 ) )                                 // sleeve corners
                  + ( _np * ( 8 * ( _nc - 1 ) + 16 ) - ( ( _closed ) ? 0 : 16 ) ) // channel corners
                ) ;
    
    sizetype cols = sizetype( _np * ( 6 * _dg - 2 ) ) ;
    
    std::vector< std::vector< Coefficient > >::size_type crows =
      std::vector< std::vector< Coefficient > >::size_type( rows ) ;
    std::vector< Bound >::size_type brows =
      std::vector< Bound >::size_type( rows ) ;

    _coefficients.resize( crows ) ;
    _bounds.resize( brows ) ;

    unsigned eqline = 0 ;

    compute_min_max_constraints( eqline ) ;
    compute_correspondence_constraints( eqline ) ;
    compute_c0continuity_constraints( eqline ) ;
    compute_c1continuity_constraints( eqline ) ;
    compute_sleeve_corners_in_channel_constraints( eqline ) ;
    compute_channel_corners_outside_sleeve_constraints( eqline ) ;
    
    error = solve_lp( rows , cols ) ;
    
    return ( error == 0 ) ;
  }

  
  // -----------------------------------------------------------------
  //
  // Private methods
  //
  // -----------------------------------------------------------------


  /**
   * \fn double BuildCurve2D::lf( double u , double b0 , double bd ) const
   *
   * \brief Computes the value of  the affine function \f$\ell\f$ at a
   * given point of the interval \f$[0,1]\f$ of the real line.
   *
   * \param u A parameter point in the interval \f$[0,1]\f$.
   * \param b0 The x- or y-coordinate  of the first control point of a
   * B&eacute;zier curve of degree d.
   * \param bd The x- or y-coordinate  of the last control point of a
   * B&eacute;zier curve of degree d.
   *
   * \return The  value of the  affine function \f$\ell\f$ at  a given
   * point of the interval \f$[0,1]\f$ of the real line.
   *
   */
  double
  BuildCurve2D::lf(
                   double u ,
                   double b0 ,
                   double bd
                  )
    const
  {
#ifdef DEBUGMODE
    assert( u >= 0 ) ;
    assert( u <= 1 ) ;
#endif

    double res = b0 + ( bd - b0 ) * u ;

    return res ;
  }


  /**
   * \fn double BuildCurve2D::h( double u ) const
   *
   * \brief Computes the value of a piecewise affine hat function at a
   * given point of the real line.
   *
   * \param u A parameter point in the real line.
   *
   * \return The value  of a piecewise linear hat function  at a given
   * point of the real line.
   *
   */
  double
  BuildCurve2D::h( double u ) const
  {
    const double onedth = 1.0 / _dg ;

    if ( u <= -onedth ) {
      return 0 ;
    }
    else if ( u <= 0 ) {
      return _dg * u + 1 ;
    }
    else if ( u <= onedth ) {
      return 1 - _dg * u ;
    }

    return 0 ;
  }


  /**
   * \fn void BuildCurve2D::compute_normal_to_lower_envelope( sizetype s , double& nx , double& ny ) const
   *
   * \brief Computes an outward normal to the s-th line segment of the
   * lower envelope of the channel.
   *
   * \param s Index of a line segment of the lower channel envelope.
   * \param nx  A reference to  the first Cartesian coordinate  of the
   * normal.
   * \param ny A  reference to the Second Cartesian  coordinate of the
   * normal.
   *
   */
  void
  BuildCurve2D::compute_normal_to_lower_envelope(
                                                 sizetype s ,
                                                 double& nx ,
                                                 double& ny
                                                )
    const
  {
#ifdef DEBUGMODE
    assert( s < ( _np * _nc ) ) ;
#endif
    sizetype t = s + 1 ;
    
    nx = _lycoords[ s ] - _lycoords[ t ] ;
    ny = _lxcoords[ t ] - _lxcoords[ s ] ;

    return ;
  }


  /**
   * \fn void BuildCurve2D::compute_normal_to_upper_envelope( sizetype s , double& nx , double& ny ) const
   *
   * \brief Computes an outward normal to the s-th line segment of the
   * upper envelope of the channel.
   *
   * \param s Index of a line segment of the upper channel envelope.
   * \param nx  A reference to  the first Cartesian coordinate  of the
   * normal.
   * \param ny A  reference to the Second Cartesian  coordinate of the
   * normal.
   *
   */
  void
  BuildCurve2D::compute_normal_to_upper_envelope(
                                                 sizetype s ,
                                                 double& nx ,
                                                 double& ny
                                                )
    const
  {
#ifdef DEBUGMODE
    assert( s < ( _np * _nc ) ) ;
#endif
    sizetype t = s + 1 ;
    
    nx = _uycoords[ t ] - _uycoords[ s ] ;
    ny = _uxcoords[ s ] - _uxcoords[ t ] ;

    return ;
  }


  /**
   * \fn unsigned BuildCurve2D::compute_control_value_column_index( unsigned p , unsigned i , unsigned v ) const
   *
   * \brief Computes  the index  of the  linear program  matrix column
   * corresponding to the x- or y-coordinate of the i-th control point
   * of the  p-th piece of  the spline curve  to be threaded  into the
   * channel.
   *
   * \param p Index of the spline curve piece.
   * \param i Index of a control point of the p-th spline curve piece.
   * \param v Index of the x- or y-coordinate of the control point.
   *
   * \return   The  index   of  the   linear  program   matrix  column
   * corresponding to the x- or y-coordinate of the i-th control point
   * of the  p-th piece of  the spline curve  to be threaded  into the
   * channel.
   *
   */
  unsigned
  BuildCurve2D::compute_control_value_column_index(
                                                   unsigned p ,
                                                   unsigned i ,
                                                   unsigned v 
                                                  )
    const
  {
#ifdef DEBUGMODE
    assert( p < _np ) ;
    assert( i <= _dg ) ;
    assert( v <= 1 ) ;
#endif

    unsigned offset = ( ( 6 * _dg ) - 2 ) * p ;

    return offset + ( 2 * i ) + v ;
  }


  /**
   * \fn unsigned BuildCurve2D::compute_second_difference_column_index( unsigned p , unsigned i , unsigned l , unsigned v ) const
   *
   * \brief Computes  the index  of the  linear program  matrix column
   * corresponding to the x- or y-coordinate  of the l-th bound of the
   * i-th second difference  of the p-th piece of the  spline curve to
   * be threaded into the channel.
   *
   * \param p Index of the spline curve piece.
   * \param i Index of the second  difference of the p-th spline curve
   * piece.
   * \param l  Index of the l-th  bound of the second  difference (0 -
   * lower bound; 1 - upper bound).
   * \param v Index of the x- or y-coordinate of the second difference
   * bound.
   *
   * \return   The  index   of  the   linear  program   matrix  column
   * corresponding to the x- or y-coordinate  of the l-th bound of the
   * i-th second difference  of the p-th piece of the  spline curve to
   * be threaded into the channel.
   *
   */
  unsigned
  BuildCurve2D::compute_second_difference_column_index(
                                                       unsigned p ,
                                                       unsigned i ,
                                                       unsigned l ,
                                                       unsigned v 
                                                      )
    const
  {
#ifdef DEBUGMODE
    assert( p < _np ) ;
    assert( i >= 1 ) ;
    assert( i < _dg ) ;
    assert( l <= 1 ) ;
    assert( v <= 1 ) ;
#endif
  
    unsigned offset = ( ( 6 * _dg ) - 2 ) * p + ( 2 * _dg + 2 ) ; 


    return offset + ( 4 * ( i - 1 ) ) + ( 2 * l ) + v ;
  }


  /**
   * \fn void BuildCurve2D::compute_min_max_constraints( unsigned& eqline )
   *
   * \brief Computes the equations defining the min-max constraints.
   *
   * \param eqline A reference to the counter of equations. 
   *
   */
  void
  BuildCurve2D::compute_min_max_constraints( unsigned& eqline )
  {
    /*
     * For each  curve p, for each  second difference i, and  for each
     * coordinate v of  the i-th second difference of  the p-th curve,
     * computes  the  three  equations corresponding  to  the  min-max
     * constraints.
     */
    for ( unsigned p = 0 ; p < _np ; p++ ) {
      for ( unsigned i = 1 ; i < _dg ; i++ ) {
        for ( unsigned v = 0 ; v < 2 ; v++ ) {
          /*
           * Get  the column  indices of  the lower  bound and  of the
           * upper bound  of the  v-th coordinate  of the  i-th second
           * difference.
           *
           */
          unsigned jl = compute_second_difference_column_index(
                                                               p ,
                                                               i ,
                                                               0 ,
                                                               v
                                                              ) ;
          
          unsigned ju = compute_second_difference_column_index(
                                                               p ,
                                                               i ,
                                                               1 ,
                                                               v
                                                              ) ;
          /*
           * Get  the  column indices  of  the  v-th coordinates  that
           * define  the i-th  second  difference of  the p-th  spline
           * curve.
           */
          unsigned c1 = compute_control_value_column_index(
                                                           p ,
                                                           i - 1 ,
                                                           v
                                                          ) ;
          unsigned c2 = compute_control_value_column_index(
                                                           p ,
                                                           i ,
                                                           v
                                                          ) ;
          unsigned c3 = compute_control_value_column_index(
                                                           p ,
                                                           i + 1 ,
                                                           v
                                                          ) ;
          
          /*
           * Set the nonzero coefficients of the next three equations.
           */
          
          /*
           * The (unknown) upper  bound of the v-th  coordinate of the
           * i-th second difference  must be greater than  or equal to
           * the value of the corresponding v-th coordinate itself.
           */         
          _coefficients[ eqline ].push_back( Coefficient( eqline , ju ,  1 ) ) ;
          _coefficients[ eqline ].push_back( Coefficient( eqline , c1 , -1 ) ) ;
          _coefficients[ eqline ].push_back( Coefficient( eqline , c2 ,  2 ) ) ;
          _coefficients[ eqline ].push_back( Coefficient( eqline , c3 , -1 ) ) ;
          _bounds[ eqline ] = Bound( Bound::GTE , 0 , eqline ) ;
          
          ++eqline ;  // increment the equation counter ;
          
          /*
           * The (unknown) upper  bound of the v-th  coordinate of the
           * i-th second difference  must be greater than  or equal to
           * zero.
           */
          _coefficients[ eqline ].push_back( Coefficient( eqline , ju ,  1 ) ) ;
          _bounds[ eqline ] = Bound( Bound::GTE , 0 , eqline ) ;
          
          ++eqline ;  // increment the equation counter ; 
          
          /*
           * The  sum  of the  upper  and  lower  bounds of  the  v-th
           * coordinate of the i-th second difference must be equal to
           * the  value of  the  v-th coordinate  of  the i-th  second
           * difference.
           */
          _coefficients[ eqline ].push_back( Coefficient( eqline , ju ,  1 ) ) ;
          _coefficients[ eqline ].push_back( Coefficient( eqline , jl ,  1 ) ) ;
          _coefficients[ eqline ].push_back( Coefficient( eqline , c1 , -1 ) ) ;
          _coefficients[ eqline ].push_back( Coefficient( eqline , c2 ,  2 ) ) ;
          _coefficients[ eqline ].push_back( Coefficient( eqline , c3 , -1 ) ) ;
          _bounds[ eqline ] = Bound( Bound::EQT , 0 , eqline ) ;
          
          ++eqline ;  // increment the equation counter ;
        }
      }
    }

    return ;
  }


  /**
   * \fn void BuildCurve2D::compute_correspondence_constraints( unsigned& eqline )
   *
   * \brief  Computes   the  equations  defining   the  correspondence
   * constraints.
   *
   * \param eqline A reference to the counter of equations. 
   *
   */
  void
  BuildCurve2D::compute_correspondence_constraints( unsigned& eqline )
  {
    /*
     * For  each curve  p, let  the coordinates  of the  first control
     * point of  the p-th  curve be  equal to  the coordinates  of the
     * middle point of the endpoints of the first c-section delimiting
     * the p-th curve.
     */
    for ( unsigned p = 0 ; p < _np ; p++ ) {
      /*
       * Get the column index of the coordinates of the first point of
       * the p-th spline curve.
       */
      unsigned jx = compute_control_value_column_index( p , 0 , 0 ) ;
      unsigned jy = compute_control_value_column_index( p , 0 , 1 ) ;
        
      /*
       * Find  the index  of  the first  c-segment  covering the  p-th
       * curve.
       */
      sizetype s = sizetype( _nc * p ) ;
        
      double x = _lxcoords[ s ] + _uxcoords[ s ] ;
      double y = _lycoords[ s ] + _uycoords[ s ] ;

      _coefficients[ eqline ].push_back( Coefficient( eqline , jx ,  2 ) ) ;
      _bounds[ eqline ] = Bound( Bound::EQT , x , eqline ) ; 
        
      ++eqline ;  // increment the equation counter ;

      _coefficients[ eqline ].push_back( Coefficient( eqline , jy ,  2 ) ) ;
      _bounds[ eqline ] = Bound( Bound::EQT , y , eqline ) ;
        
      ++eqline ;  // increment the equation counter ;
    }
    
    if ( !_closed ) {
      /*
       * Get the column  index of the coordinates of  the last control
       * point of the last spline curve.
       */
      unsigned jx = compute_control_value_column_index(
                                                       _np - 1 ,
                                                       _dg ,
                                                       0
                                                      ) ;
      unsigned jy = compute_control_value_column_index(
                                                       _np - 1 ,
                                                       _dg ,
                                                       1
                                                      ) ;
      
      /*
       * Find the  index of  the last  c-segments delimiting  the last
       * curve.
       */
      sizetype s = sizetype( _nc * _np ) ;
        
      double x = _lxcoords[ s ] + _uxcoords[ s ] ;
      double y = _lycoords[ s ] + _uycoords[ s ] ;
        
      _coefficients[ eqline ].push_back( Coefficient( eqline , jx ,  2 ) ) ;
      _bounds[ eqline ] = Bound( Bound::EQT , x , eqline ) ; 
        
      ++eqline ;  // increment the equation counter ;
        
      _coefficients[ eqline ].push_back( Coefficient( eqline , jy ,  2 ) ) ;
      _bounds[ eqline ] = Bound( Bound::EQT , y , eqline ) ; 
        
      ++eqline ;  // increment the equation counter ;
    }
    
    return ;
  }


  /**
   * \fn void BuildCurve2D::compute_c0continuity_constraints( unsigned& eqline )
   *
   * \brief   Computes  the   equations  defining   the  C0-continuity
   * constraints.
   *
   * \param eqline A reference to the counter of equations. 
   *
   */
  void
  BuildCurve2D::compute_c0continuity_constraints( unsigned& eqline )
  {
    /*
     * For each curve p, with p less  than the number of pieces of the
     * spline curve, make the last control point of the p-th piece the
     * same as  the first control point  of the (p+1)-th piece  of the
     * curve.
     */
    for ( unsigned p = 0 ; p < ( _np - 1 ) ; p++ ) {
      for ( unsigned v = 0 ; v < 2 ; v++ ) {
        /*
         * Get the  column index  of the v-th  coordinate of  the last
         * control point of  the p-th piece and the index  of the v-th
         * coordinate  of  the first  control  point  of the  (p+1)-th
         * piece.
         */
        unsigned jd = compute_control_value_column_index(
                                                         p ,
                                                         _dg ,
                                                         v
                                                        ) ;
        
        unsigned j0 = compute_control_value_column_index(
                                                         p + 1 ,
                                                         0 ,
                                                         v
                                                        ) ;
        
        _coefficients[ eqline ].push_back( Coefficient( eqline , jd ,  1 ) ) ;
        _coefficients[ eqline ].push_back( Coefficient( eqline , j0 , -1 ) ) ;
        _bounds[ eqline ] = Bound( Bound::EQT , 0 , eqline ) ;
        
        ++eqline ;  // increment the equation counter ;
      }
    }
    
    if ( _closed ) {
      for ( unsigned v = 0 ; v < 2 ; v++ ) {
        /*
         * Get the  column index  of the v-th  coordinate of  the last
         * control point of  the p-th piece and the index  of the v-th
         * coordinate of the first control point of the first piece.
         */
        unsigned jd = compute_control_value_column_index(
                                                         _np - 1 ,
                                                         _dg ,
                                                         v
                                                        ) ;
        
        unsigned j0 = compute_control_value_column_index(
                                                         0 ,
                                                         0 ,
                                                         v
                                                        ) ;
        
        _coefficients[ eqline ].push_back( Coefficient( eqline , jd ,  1 ) ) ;
        _coefficients[ eqline ].push_back( Coefficient( eqline , j0 , -1 ) ) ;
        _bounds[ eqline ] = Bound( Bound::EQT , 0 , eqline ) ;
        
        ++eqline ;  // increment the equation counter ;          
      }
    }
    
    return ;
  }


  /**
   * \fn void BuildCurve2D::compute_c1continuity_constraints( unsigned& eqline )
   *
   * \brief   Computes  the   equations  defining   the  C1-continuity
   * constraints.
   *
   * \param eqline A reference to the counter of equations.
   *
   */
  void
  BuildCurve2D::compute_c1continuity_constraints( unsigned& eqline )
  {
    /*
     * For each curve p, with p less  than the number of pieces of the
     * spline curve,  make the  last control point  of the  p-th piece
     * equal to the midpoint of the line segment defined by the before
     * the last  control point of p-th  piece and the second  point of
     * the (p+1)-th piece of the curve.
     */
    for ( unsigned p = 0 ; p < ( _np - 1 ) ; p++ ) {
      for ( unsigned v = 0 ; v < 2 ; v++ ) {
        /*
         * Get the column indices of  the v-th coordinates of the last
         * control  point  of the  p-th  piece,  the before  the  last
         * control point  of the  p-th piece,  and the  second control
         * point of the (p+1)-th piece of the curve.
         */
        unsigned jd = compute_control_value_column_index(
                                                         p ,
                                                         _dg ,
                                                         v
                                                         ) ;
        
        unsigned jp = compute_control_value_column_index(
                                                         p ,
                                                         _dg - 1 ,
                                                         v
                                                         ) ;
        
        unsigned j2 = compute_control_value_column_index(
                                                         p + 1 ,
                                                         1 ,
                                                         v
                                                         ) ;
        
        _coefficients[ eqline ].push_back( Coefficient( eqline , jd ,  2 ) ) ;
        _coefficients[ eqline ].push_back( Coefficient( eqline , jp , -1 ) ) ;
        _coefficients[ eqline ].push_back( Coefficient( eqline , j2 , -1 ) ) ;
        _bounds[ eqline ] = Bound( Bound::EQT , 0 , eqline ) ;
        
        ++eqline ;  // increment the equation counter ;
      }
    }
    
    if ( _closed ) {
      for ( unsigned v = 0 ; v < 2 ; v++ ) {
        /*
         * Get the column indices of  the v-th coordinates of the last
         * control  point  of the  last  piece,  the before  the  last
         * control point  of the  last piece,  and the  second control
         * point of the first piece of the spline.
         */
        unsigned jd = compute_control_value_column_index(
                                                         _np - 1 ,
                                                         _dg ,
                                                         v
                                                        ) ;
        
        unsigned jp = compute_control_value_column_index(
                                                         _np - 1 ,
                                                         _dg - 1 ,
                                                         v
                                                        ) ;
        
        unsigned j2 = compute_control_value_column_index(
                                                         0 ,
                                                         1 ,
                                                         v
                                                        ) ;
        
        _coefficients[ eqline ].push_back( Coefficient( eqline , jd ,  2 ) ) ;
        _coefficients[ eqline ].push_back( Coefficient( eqline , jp , -1 ) ) ;
        _coefficients[ eqline ].push_back( Coefficient( eqline , j2 , -1 ) ) ;
        _bounds[ eqline ] = Bound( Bound::EQT , 0 , eqline ) ;
        
        ++eqline ;  // increment the equation counter ;
      }
    }
    
    return ;
  }


  /**
   * \fn void BuildCurve2D::compute_sleeve_corners_in_channel_constraints( unsigned& eqline )
   *
   * \brief  Computes  the  equations defining  the  constraints  that
   * ensure that the corners of the sleeves are inside the channel.
   *
   * \param eqline A reference to the counter of equations. 
   *
   */
  void
  BuildCurve2D::compute_sleeve_corners_in_channel_constraints( unsigned& eqline )
  {
    /*
     * Compute outward normals to the line segments of the c-segments.
     */
    std::vector< std::vector< double > > nl( _np * _nc , std::vector< double >( 2 , 0 ) ) ;
    std::vector< std::vector< double > > nu( _np * _nc , std::vector< double >( 2 , 0 ) ) ;
    for ( sizetype c = 0 ; c < sizetype( _np * _nc ) ; c++ ) {
      compute_normal_to_lower_envelope( c , nl[ c ][ 0 ] , nl[ c ][ 1 ] ) ;
      compute_normal_to_upper_envelope( c , nu[ c ][ 0 ] , nu[ c ][ 1 ] ) ;
    }
    
    /*
     * For each  curve p, each  sleeve corner  of the p-th  curve, and
     * each line segment  bounding the sleeve corner  (i.e., the lower
     * and upper  line segment of  a c-segment), compute  a constraint
     * that ensures that  the sleeve corner is on the  correct side of
     * the segment.
     */
    for ( unsigned p = 0 ; p < _np ; p++ ) {
      for ( unsigned s = 0 ; s <= _dg ; s++ ) {
        /*
         * Find the  index \e c of  the upper and lower  line segments
         * that bound the s-th corner of the sleeve of the p-th curve.
         */
        sizetype c = sizetype( floor( s * ( double( _nc ) / _dg ) ) ) ;
        if ( s == _dg ) {
          --c ;
        }
        c += sizetype( ( p * _nc ) ) ;
        
        double t = double( s ) / _dg ;

        /*
         * Compute  the column  indices of  the linear  program matrix
         * corresponding  to the  second differences  and the  control
         * points involved in the constraints associated with the c-th
         * corner.
         */
        std::vector< std::vector< unsigned > > cp( 2 , std::vector< unsigned >( 2 , 0 ) ) ;
        
        cp[ 0 ][ 0 ] = compute_control_value_column_index(
                                                          p ,
                                                          0 ,
                                                          0
                                                         ) ;
        cp[ 0 ][ 1 ] = compute_control_value_column_index(
                                                          p ,
                                                          0 ,
                                                          1
                                                         ) ;
        cp[ 1 ][ 0 ] = compute_control_value_column_index(
                                                          p ,
                                                          _dg ,
                                                          0
                                                          ) ;
        cp[ 1 ][ 1 ] = compute_control_value_column_index(
                                                          p ,
                                                          _dg ,
                                                          1
                                                          ) ;

        std::vector< std::vector< std::vector< unsigned > > > sd(
                                                               _dg - 1 ,
                                                               std::vector< std::vector< unsigned > >
                                                               (
                                                                 2 ,
                                                                 std::vector< unsigned >( 2 , 0 )
                                                               )
                                                              ) ;
        
        for ( unsigned j = 1 ; j < _dg ; j++ ) {
          for ( unsigned l = 0 ; l < 2 ; l++ ) {
            for ( unsigned v = 0 ; v < 2 ; v++ ) {
              /*
               * Compute  the column  indices  of  the linear  program
               * matrix  corresponding to  the  values  of the  second
               * differences in the constraints.
               */
              sd[ j - 1 ][ l ][ v ] = compute_second_difference_column_index(
                                                                             p ,
                                                                             j ,
                                                                             l ,
                                                                             v
                                                                            ) ;
            }
          }
        }

        /*
         * Compute the constraints corresponding to the lower envelope of the sleeve.
         */
        
        for ( unsigned j = 1 ; j < _dg ; j++ ) {
          double du ;
          double dl ;
	  try {
	    du = _tf->alower( j , t ) ;
	    dl = _tf->aupper( j , t ) ;
	  }
	  catch ( const ExceptionObject& xpt ) {
	    treat_exception( xpt ) ;
	    exit( EXIT_FAILURE ) ;
	  }
          
          for ( sizetype v = 0 ; v < 2 ; v++ ) {
            double temp ;
            temp = dl * nl[ c ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline     ].push_back( Coefficient( eqline     , sd[ sizetype( j ) - 1 ][ 0 ][ sizetype( v ) ] ,  temp ) ) ;
            }
            
            temp = du * nl[ c ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline     ].push_back( Coefficient( eqline     , sd[ sizetype( j ) - 1 ][ 1 ][ sizetype( v ) ] ,  temp ) ) ;
            }
            
            temp = dl * nu[ c ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , sd[ sizetype( j ) - 1 ][ 0 ][ sizetype( v ) ] ,  temp ) ) ;
            }
            
            temp = du * nu[ c ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , sd[ sizetype( j ) - 1 ][ 1 ][ sizetype( v ) ] ,  temp ) ) ;
            }
          }
        }
        
        for ( sizetype v = 0 ; v < 2 ; v++ ) {
          double temp ;
          temp = ( 1 - t ) * nl[ c ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline     ].push_back( Coefficient( eqline     , cp[ 0 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = (     t ) * nl[ c ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline     ].push_back( Coefficient( eqline     , cp[ 1 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = ( 1 - t ) * nu[ c ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , cp[ 0 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = (     t ) * nu[ c ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , cp[ 1 ][ sizetype( v ) ] , temp ) ) ;
          }
        }
        
        _bounds[ eqline     ] = Bound( Bound::LTE , _lxcoords[ c ] * nl[ c ][ 0 ] + _lycoords[ c ] * nl[ c ][ 1 ] , eqline     ) ;
        _bounds[ eqline + 1 ] = Bound( Bound::LTE , _uxcoords[ c ] * nu[ c ][ 0 ] + _uycoords[ c ] * nu[ c ][ 1 ] , eqline + 1 ) ;
        
        /*
         * Increment equation counter
         */
        eqline += 2 ;
        
        /*
         * Compute the constraints corresponding to the upper envelope of the sleeve.
         */
        
        for ( unsigned j = 1 ; j < _dg ; j++ ) {
          double du ;
          double dl ;
	  try {
	    du = _tf->aupper( j , t ) ;
	    dl = _tf->alower( j , t ) ;
	  }
	  catch ( const ExceptionObject& xpt ) {
	    treat_exception( xpt ) ;
	    exit( EXIT_FAILURE ) ;
	  }
          
          for ( sizetype v = 0 ; v < 2 ; v++ ) {
            double temp ;
            temp = dl * nl[ c ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline     ].push_back( Coefficient( eqline     , sd[ sizetype( j ) - 1 ][ 0 ][ sizetype( v ) ] , temp ) ) ;
            }
            
            temp = du * nl[ c ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline     ].push_back( Coefficient( eqline     , sd[ sizetype( j ) - 1 ][ 1 ][ sizetype( v ) ] , temp ) ) ;
            }
            
            temp = dl * nu[ c ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , sd[ sizetype( j ) - 1 ][ 0 ][ sizetype( v ) ] , temp ) ) ;
            }
            
            temp = du * nu[ c ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , sd[ sizetype( j ) - 1 ][ 1 ][ sizetype( v ) ] , temp ) ) ;
            }
          }
        }
        
        for ( sizetype v = 0 ; v < 2 ; v++ ) {
          double temp ;
          temp = ( 1 - t ) * nl[ c ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline     ].push_back( Coefficient( eqline     , cp[ 0 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = (     t ) * nl[ c ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline     ].push_back( Coefficient( eqline     , cp[ 1 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = ( 1 - t ) * nu[ c ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , cp[ 0 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = (     t ) * nu[ c ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , cp[ 1 ][ sizetype( v ) ] , temp ) ) ;
          }
        }
        
        _bounds[ eqline     ] = Bound( Bound::LTE , _lxcoords[ c ] * nl[ c ][ 0 ] + _lycoords[ c ] * nl[ c ][ 1 ] , eqline     ) ;
        _bounds[ eqline + 1 ] = Bound( Bound::LTE , _uxcoords[ c ] * nu[ c ][ 0 ] + _uycoords[ c ] * nu[ c ][ 1 ] , eqline + 1 ) ;
        
        /*
         * Increment equation counter
         */
        eqline += 2 ;
      }
    }

    return ;
  }


  /**
   * \fn void BuildCurve2D::compute_channel_corners_outside_sleeve_constraints( unsigned& eqline )
   *
   * \brief  Computes  the  equations defining  the  constraints  that
   * ensure that  the corners of  the channel are located  outside the
   * sleeve.
   *
   * \param eqline A reference to the counter of equations. 
   *
   */
  void
  BuildCurve2D::compute_channel_corners_outside_sleeve_constraints( unsigned& eqline )
  {
    /*
     * Compute outward normals to the line segments of the c-segments.
     */
    std::vector< std::vector< double > > nl( _np * _nc , std::vector< double >( 2 , 0 ) ) ;
    std::vector< std::vector< double > > nu( _np * _nc , std::vector< double >( 2 , 0 ) ) ;
    if ( ( _np > 1 ) || ( _nc > 1 ) ) {
      for ( sizetype c = 0 ; c < sizetype( _np * _nc ) ; c++ ) {
        compute_normal_to_lower_envelope( c , nl[ c ][ 0 ] , nl[ c ][ 1 ] ) ;
        compute_normal_to_upper_envelope( c , nu[ c ][ 0 ] , nu[ c ][ 1 ] ) ;
      }
    }

    /*
     * For each  curve p, for  each channel  corner that belongs  to a
     * c-segment delimiting the p-th curve piece, compute a constraint
     * that ensures that  the channel corner is outside  the sleeve of
     * the p-th piece.
     */
    for ( unsigned p = 0 ; p < _np ; p++ ) {
      for ( unsigned c = 0 ; c <= _nc ; c++ ) {
        /*
         * Check whether  the current constraint is  equivalent to the
         * sleeve inside channel constraint. If so, there is no reason
         * to create another equation, which would otherwise slow down
         * computations.
         */
        if ( ( c == 0 ) && ( p == 0 ) && ( !_closed ) ) {
          continue ;
        }
        
        if ( ( c == _nc ) && ( p == ( _np - 1 ) ) && ( !_closed ) ) {
          continue ;
        }
        
        /*
         * Find the  index \e s  of the  nearest sleeve corner  at the
         * c-section containing the  channel corner or to  the left of
         * it.
         */
        sizetype s = sizetype( ceil( c * ( double( _dg ) / _nc ) ) ) ;
        
        /*
         * Find the  index \e t  of the  nearest sleeve corner  at the
         * c-section containing the channel corner  or to the right of
         * it.
         */
        sizetype t = ( s == 0 ) ? s : s - 1 ;
        
        /*
         * Compute the  parameter value  that gives the  e-piece point
         * that matches the channel corner in a parametrization of the
         * c-piece.
         */
        double u = s - c * ( double( _dg ) / _nc ) ;
        
        /*
         * Copmpute the parameter values of a point in the e-piece.
         */
        double sp = double( s ) / _dg ;
        double tp = double( t ) / _dg ;
        
        /*
         * Select  the  normals to  the  line  segments of  the  lower
         * envelope incident  to the channel  corner \e c of  the p-th
         * piece.
         */
        sizetype cl = ( ( p * _nc ) + c ) % ( _np * _nc ) ;
        sizetype cr = ( cl == 0 ) ? ( _np * _nc ) - 1 : cl - 1 ;
        
        /*
         * Compute  the column  indices of  the linear  program matrix
         * corresponding  to the  second differences  and the  control
         * points involved in the constraints associated with the c-th
         * corner.
         */
        std::vector< std::vector< unsigned > > cp( 2 , std::vector< unsigned >( 2 , 0 ) ) ;
        
        cp[ 0 ][ 0 ] = compute_control_value_column_index(
                                                          p ,
                                                          0 ,
                                                          0
                                                         ) ;
        cp[ 0 ][ 1 ] = compute_control_value_column_index(
                                                          p ,
                                                          0 ,
                                                          1
                                                         ) ;
        cp[ 1 ][ 0 ] = compute_control_value_column_index(
                                                          p ,
                                                          _dg ,
                                                          0
                                                          ) ;
        cp[ 1 ][ 1 ] = compute_control_value_column_index(
                                                          p ,
                                                          _dg ,
                                                          1
                                                          ) ;
        
        std::vector< std::vector< std::vector< unsigned > > > sd(
                                                                _dg - 1 ,
                                                                std::vector< std::vector< unsigned > >(
                                                                  2 ,
                                                                  std::vector< unsigned >( 2 , 0 )
                                                                )
                                                              ) ;
        
        for ( unsigned j = 1 ; j < _dg ; j++ ) {
          for ( unsigned l = 0 ; l < 2 ; l++ ) {
            for ( unsigned v = 0 ; v < 2 ; v++ ) {
              /*
               * Compute  the column  indices  of  the linear  program
               * matrix  corresponding to  the  values  of the  second
               * differences in the constraints.
               */
              sd[ j - 1 ][ l ][ v ] = compute_second_difference_column_index(
                                                                             p ,
                                                                             j ,
                                                                             l ,
                                                                             v
                                                                            ) ;
            }
          }
        }

        
        /*
         * Constraints related to the sleeve lower envelope.
         */
        
        for ( unsigned j = 1 ; j < _dg ; j++ ) {
	  double du ;
          double dl ;
	  try {
	    du = ( u * _tf->alower( j , tp ) ) + ( ( 1 - u ) * _tf->alower( j , sp ) ) ;
	    dl = ( u * _tf->aupper( j , tp ) ) + ( ( 1 - u ) * _tf->aupper( j , sp ) ) ;
	  }
	  catch ( const ExceptionObject& xpt ) {
	    treat_exception( xpt ) ;
	    exit( EXIT_FAILURE ) ;
	  }

          for ( sizetype v = 0 ; v < 2 ; v++ ) {
            double temp ;
            temp = dl * nl[ cr ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline     ].push_back( Coefficient( eqline     , sd[ sizetype( j ) - 1 ][ 0 ][ sizetype( v ) ] , temp  ) ) ;
            }
            
            temp = du * nl[ cr ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline     ].push_back( Coefficient( eqline     , sd[ sizetype( j ) - 1 ][ 1 ][ sizetype( v ) ] , temp  ) ) ;
            }
            
            temp = dl * nl[ cl ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , sd[ sizetype( j ) - 1 ][ 0 ][ sizetype( v ) ] , temp  ) ) ;
            }
            
            temp = du * nl[ cl ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , sd[ sizetype( j ) - 1 ][ 1 ][ sizetype( v ) ] , temp  ) ) ;
            }
            
            temp = dl * nu[ cr ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 2 ].push_back( Coefficient( eqline + 2 , sd[ sizetype( j ) - 1 ][ 0 ][ sizetype( v ) ] , temp  ) ) ;
            }
            
            temp = du * nu[ cr ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 2 ].push_back( Coefficient( eqline + 2 , sd[ sizetype( j ) - 1 ][ 1 ][ sizetype( v ) ] , temp  ) ) ;
            }
            
            temp = dl * nu[ cl ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 3 ].push_back( Coefficient( eqline + 3 , sd[ sizetype( j ) - 1 ][ 0 ][ sizetype( v ) ] , temp ) ) ;
            }
            
            temp = du * nu[ cl ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 3 ].push_back( Coefficient( eqline + 3 , sd[ sizetype( j ) - 1 ][ 1 ][ sizetype( v ) ] , temp ) ) ;
            }
          }
        }
        
        double p0 = ( ( 1 - u ) * ( 1 - sp ) ) + ( u * ( 1 - tp ) )  ;
        double pd = ( ( 1 - u ) * (     sp ) ) + ( u * (     tp ) )  ;
        
        for ( sizetype v = 0 ; v < 2 ; v++ ) {
          double temp ;
          temp = p0 * nl[ cr ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline     ].push_back( Coefficient( eqline     , cp[ 0 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = pd * nl[ cr ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline     ].push_back( Coefficient( eqline     , cp[ 1 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = p0 * nl[ cl ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , cp[ 0 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = pd * nl[ cl ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , cp[ 1 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = p0 * nu[ cr ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 2 ].push_back( Coefficient( eqline + 2 , cp[ 0 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = pd * nu[ cr ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 2 ].push_back( Coefficient( eqline + 2 , cp[ 1 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = p0 * nu[ cl ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 3 ].push_back( Coefficient( eqline + 3 , cp[ 0 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = pd * nu[ cl ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 3 ].push_back( Coefficient( eqline + 3 , cp[ 1 ][ sizetype( v ) ] , temp ) ) ;
          }
        }
        
        _bounds[ eqline     ] = Bound( Bound::LTE , _lxcoords[ cl ] * nl[ cr ][ 0 ] + _lycoords[ cl ] * nl[ cr ][ 1 ] , eqline     ) ;
        _bounds[ eqline + 1 ] = Bound( Bound::LTE , _lxcoords[ cl ] * nl[ cl ][ 0 ] + _lycoords[ cl ] * nl[ cl ][ 1 ] , eqline + 1 ) ;
        
        _bounds[ eqline + 2 ] = Bound( Bound::LTE , _uxcoords[ cl ] * nu[ cr ][ 0 ] + _uycoords[ cl ] * nu[ cr ][ 1 ] , eqline + 2 ) ;
        _bounds[ eqline + 3 ] = Bound( Bound::LTE , _uxcoords[ cl ] * nu[ cl ][ 0 ] + _uycoords[ cl ] * nu[ cl ][ 1 ] , eqline + 3 ) ;
        
        /*
         * Increment equation counter
         */
        eqline += 4 ;
        
        /*
         * Constraints related to the sleeve upper envelope.
         */
        
        for ( unsigned j = 1 ; j < _dg ; j++ ) {
          double du ;
          double dl ; 
	  try {
	    du = ( u * _tf->aupper( j , tp ) ) + ( ( 1 - u ) * _tf->aupper( j , sp ) ) ;
	    dl = ( u * _tf->alower( j , tp ) ) + ( ( 1 - u ) * _tf->alower( j , sp ) ) ;
	  }
	  catch ( const ExceptionObject& xpt ) {
	    treat_exception( xpt ) ;
	    exit( EXIT_FAILURE ) ;
	  }
          
          for ( sizetype v = 0 ; v < 2 ; v++ ) {
            double temp ;
            temp = dl * nl[ cr ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline     ].push_back( Coefficient( eqline     , sd[ sizetype( j ) - 1 ][ 0 ][ sizetype( v ) ] , temp ) ) ;
            }
            
            temp = du * nl[ cr ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline     ].push_back( Coefficient( eqline     , sd[ sizetype( j ) - 1 ][ 1 ][ sizetype( v ) ] , temp ) ) ;
            }
            
            temp = dl * nl[ cl ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , sd[ sizetype( j ) - 1 ][ 0 ][ sizetype( v ) ] , temp ) ) ;
            }
            
            temp = du * nl[ cl ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , sd[ sizetype( j ) - 1 ][ 1 ][ sizetype( v ) ] , temp ) ) ;
            }
            
            temp = dl * nu[ cr ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 2 ].push_back( Coefficient( eqline + 2 , sd[ sizetype( j ) - 1 ][ 0 ][ sizetype( v ) ] , temp ) ) ;
            }
            
            temp = du * nu[ cr ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 2 ].push_back( Coefficient( eqline + 2 , sd[ sizetype( j ) - 1 ][ 1 ][ sizetype( v ) ] , temp ) ) ;
            }
            
            temp = dl * nu[ cl ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 3 ].push_back( Coefficient( eqline + 3 , sd[ sizetype( j ) - 1 ][ 0 ][ sizetype( v ) ] , temp ) ) ;
            }
            
            temp = du * nu[ cl ][ sizetype( v ) ] ;
            if ( fabs( temp ) > 0 ) {
              _coefficients[ eqline + 3 ].push_back( Coefficient( eqline + 3 , sd[ sizetype( j ) - 1 ][ 1 ][ sizetype( v ) ] , temp ) ) ;
            }
          }
        }
        
        for ( sizetype v = 0 ; v < 2 ; v++ ) {
          double temp ;
          temp = p0 * nl[ cr ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline     ].push_back( Coefficient( eqline     , cp[ 0 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = pd * nl[ cr ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline     ].push_back( Coefficient( eqline     , cp[ 1 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = p0 * nl[ cl ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , cp[ 0 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = pd * nl[ cl ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 1 ].push_back( Coefficient( eqline + 1 , cp[ 1 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = p0 * nu[ cr ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 2 ].push_back( Coefficient( eqline + 2 , cp[ 0 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = pd * nu[ cr ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 2 ].push_back( Coefficient( eqline + 2 , cp[ 1 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = p0 * nu[ cl ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 3 ].push_back( Coefficient( eqline + 3 , cp[ 0 ][ sizetype( v ) ] , temp ) ) ;
          }
          
          temp = pd * nu[ cl ][ sizetype( v ) ] ;
          if ( fabs( temp ) > 0 ) {
            _coefficients[ eqline + 3 ].push_back( Coefficient( eqline + 3 , cp[ 1 ][ sizetype( v ) ] , temp ) ) ;
          }
        }
        
        _bounds[ eqline     ] = Bound( Bound::LTE , _lxcoords[ cl ] * nl[ cr ][ 0 ] + _lycoords[ cl ] * nl[ cr ][ 1 ] , eqline     ) ;
        _bounds[ eqline + 1 ] = Bound( Bound::LTE , _lxcoords[ cl ] * nl[ cl ][ 0 ] + _lycoords[ cl ] * nl[ cl ][ 1 ] , eqline + 1 ) ;

        _bounds[ eqline + 2 ] = Bound( Bound::LTE , _uxcoords[ cl ] * nu[ cr ][ 0 ] + _uycoords[ cl ] * nu[ cr ][ 1 ] , eqline + 2 ) ;
        _bounds[ eqline + 3 ] = Bound( Bound::LTE , _uxcoords[ cl ] * nu[ cl ][ 0 ] + _uycoords[ cl ] * nu[ cl ][ 1 ] , eqline + 3 ) ;
        
        /*
         * Increment equation counter
         */
        eqline += 4 ;
      }
    }
    
    return ;
  }


  /**
   * \fn int BuildCurve2D::solve_lp( sizetype rows , sizetype cols )
   *
   * \brief  Solves the  linear program  corresponding to  the channel
   * problem.
   *
   * \param rows The number of constraints of the linear program.
   * \param cols The number of unknowns of the linear program.
   *
   * \return The code returned by the LP solver to indicate the status
   * of the computation of the solution of the linear program.
   */
  int
  BuildCurve2D::solve_lp(
                         sizetype rows ,
                         sizetype cols
                        )
  {
    /*
     * Create the LP problem.
     */
    glp_prob* lp = glp_create_prob() ;
    
    /*
     * Set up the number of constraints and structural variables.
     */
    glp_add_rows( lp , int( rows ) ) ;
    glp_add_cols( lp , int( cols ) ) ;
    
    /*
     * Set the problem as a minimization one.
     */
    glp_set_obj_dir( lp , GLP_MIN ) ;
    
    /*
     * Set up the constraints of the problem.
     */
    set_up_lp_constraints( lp ) ;
    
    /*
     * Define bounds on the structural variables of the problem.
     */
    set_up_structural_variables( lp ) ;
    
    /*
     * Define objective function.
     */
    set_up_objective_function( lp ) ; 
    
    /*
     * Set parameters of the solver.
     */
    glp_smcp param ;
    glp_init_smcp( &param ) ;
    
    param.msg_lev  = GLP_MSG_OFF ;
    param.presolve = GLP_ON ;
    
    /*
     * Call the solver.
     */
    
    int res = glp_simplex( lp , &param ) ;

    if ( res == 0 ) {
      /*
       * Get the solver result information.
       */
      get_lp_solver_result_information( lp ) ;
    }
      
    /*
     * Release memory held by the solver.
     */
    glp_delete_prob( lp ) ;
    
    return res ;
  }


  /**
   * \fn void BuildCurve2D::set_up_lp_constraints( glp_prob* lp ) const
   *
   * \brief Assemble the matrix of  constraints of the linear program,
   * and define  the type (equality  or inequality) and bounds  on the
   * constraints.
   *
   * \param lp A pointer to the instance of the LP program.
   *
   */
  void
  BuildCurve2D::set_up_lp_constraints( glp_prob* lp ) const
  {
    /*
     * Set up the bounds on the constraints of the problem.
     */
    
    for ( std::vector< Bound >::size_type j = 0 ; j < _bounds.size() ; j++ ) {
#ifdef DEBUGMODE
      assert( unsigned( j ) == _bounds[ j ].get_row() ) ;
#endif
      
      int i = int( _bounds[ j ].get_row() + 1 ) ;
      
      std::stringstream ss ( std::stringstream::in | std::stringstream::out ) ;
      ss << "c" << i ;
      glp_set_row_name( lp , i , ss.str().c_str() ) ;
      
      double val = _bounds[ j ].get_value() ;
      if ( _bounds[ j ].get_type() == Bound::LTE ) {
        glp_set_row_bnds( lp , i , GLP_UP ,   0 , val ) ;
      }
      else if ( _bounds[ j ].get_type() == Bound::GTE ) {
        glp_set_row_bnds( lp , i , GLP_LO , val ,   0 ) ;
      }
      else {
        glp_set_row_bnds( lp , i , GLP_FX , val , val ) ;
      }
    }
    
    
    /*
     * Obtain the coefficients of the constraints of the problem.
     */
    
    std::vector< int    > ia ; ia.push_back( 0 ) ;  // GLPK starts indexing array \e ia at 1
    std::vector< int    > ja ; ja.push_back( 0 ) ;  // GLPK starts indexing array \e ja at 1
    std::vector< double > ar ; ar.push_back( 0 ) ;  // GLPK starts indexing array \e ar at 1
    
    int h = 0 ;
    for ( std::vector< std::vector< Coefficient > >::size_type j = 0 ; j < _coefficients.size() ; j++ ) {
      for ( std::vector< Coefficient >::size_type k = 0 ; k < _coefficients[ j ].size() ; k++ ) {
#ifdef DEBUGMODE
        assert( _coefficients[ j ][ k ].get_row() == unsigned( j ) ) ;
#endif
        ia.push_back( int( _coefficients[ j ][ k ].get_row() + 1 ) ) ;
        ja.push_back( int( _coefficients[ j ][ k ].get_col() + 1 ) ) ;
        ar.push_back( _coefficients[ j ][ k ].get_value() ) ;
        ++h ;
      }
    }
    
    glp_load_matrix(
                    lp ,
                    h ,
                    &ia[ 0 ] ,
                    &ja[ 0 ] ,
                    &ar[ 0 ]
                    ) ;
    
    return ;
  }


  /**
   * \fn void BuildCurve2D::set_up_structural_variables( glp_prob* lp ) const
   *
   * \brief  Define  lower  and/or  upper  bounds  on  the  structural
   * variables  of the  linear  program corresponding  to the  channel
   * problem.
   *
   * \param lp A pointer to the instance of the LP program.
   *
   */
  void
  BuildCurve2D::set_up_structural_variables( glp_prob* lp ) const
  {      
    for ( unsigned p = 0 ; p < _np ; p++ ) {
      for ( unsigned k = 1 ; k < _dg ; k++ ) {
        for ( unsigned l = 0 ; l < 2 ; l++ ) {
          for ( unsigned v = 0 ; v < 2 ; v++ ) {
            unsigned c = compute_second_difference_column_index(
                                                                  p ,
                                                                  k ,
                                                                  l ,
                                                                  v
                                                                  ) ;
              
            if ( l == 0 ) {
              std::stringstream ss ( std::stringstream::in | std::stringstream::out ) ;
              if ( v == 0 ) {
                ss << "mx" << ( ( _dg - 1 ) * p ) + k ;
              }
              else {
                ss << "my" << ( ( _dg - 1 ) * p ) + k ;
              }
              glp_set_col_name( lp , int( c ) + 1 , ss.str().c_str() ) ;
              glp_set_col_bnds( lp , int( c ) + 1 , GLP_UP , 0 , 0 ) ;
            }
            else {
              std::stringstream ss ( std::stringstream::in | std::stringstream::out ) ;
              if ( v == 0 ) {
                ss << "px" << ( ( _dg - 1 ) * p ) + k ;
              }
              else {
                ss << "py" << ( ( _dg - 1 ) * p ) + k ;
              }
              glp_set_col_name( lp , int( c ) + 1 , ss.str().c_str() ) ;
              glp_set_col_bnds( lp , int( c ) + 1 , GLP_LO , 0 , 0 ) ;
            }
          }
        }
      }
    }

    for ( unsigned p = 0 ; p < _np ; p++ ) {
      for ( unsigned k = 0 ; k <= _dg ; k++ ) {
        for ( unsigned v = 0 ; v < 2 ; v++ ) {
          unsigned c = compute_control_value_column_index(
                                                          p ,
                                                          k ,
                                                          v
                                                         ) ;
            
          std::stringstream ss ( std::stringstream::in | std::stringstream::out ) ;
          if ( v == 0 ) {
            ss << "x" << ( ( _dg + 1 ) * p ) + k + 1 ;
          }
          else {
            ss << "y" << ( ( _dg + 1 ) * p ) + k + 1 ;
          }
          glp_set_col_name( lp , int( c ) + 1 , ss.str().c_str() ) ;
          glp_set_col_bnds( lp , int( c ) + 1 , GLP_FR , 0 , 0 ) ;
        }
      }
    }
    
    return ;
  }


  /**
   * \fn void BuildCurve2D::set_up_objective_function( glp_prob* lp ) const
   *
   * \brief  Define  the  objective  function of  the  linear  program
   * corresponding  to the  channel problem,  which is  a minimization
   * problem.
   *
   * \param lp A pointer to the instance of the LP program.
   *
   */
  void
  BuildCurve2D::set_up_objective_function( glp_prob* lp ) const
  {
    for ( unsigned p = 0 ; p < _np ; p++ ) {
      for ( unsigned k = 1 ; k < _dg ; k++ ) {
        for ( unsigned l = 0 ; l < 2 ; l++ ) {
          for ( unsigned v = 0 ; v < 2 ; v++ ) {
            unsigned c = compute_second_difference_column_index(
                                                                p ,
                                                                k ,
                                                                l ,
                                                                v
                                                               ) ;
    
            if ( l == 0 ) {
              glp_set_obj_coef( lp , int( c ) + 1 , -1 ) ;
            }
            else {
              glp_set_obj_coef( lp , int( c ) + 1 ,  1 ) ;
            }
          }
        }
      }
    }

    return ;
  }


  /**
   * \fn void BuildCurve2D::get_lp_solver_result_information( glp_prob* lp )
   *
   * \brief Obtain the  optimal values found by the LP  solver for the
   * structural values of the  linear programming corresponding to the
   * channel problem.
   *
   * \param lp A pointer to the instance of the LP program.
   *
   */
  void
  BuildCurve2D::get_lp_solver_result_information( glp_prob* lp )
  {
    /*
     * Obtain the control points of the spline curve.
     */
    for ( unsigned p = 0 ; p < _np ; p++ ) {
      for ( unsigned k = 0 ; k <= _dg ; k++ ) {
        for ( unsigned v = 0 ; v < 2 ; v++ ) {
          unsigned c = compute_control_value_column_index(
                                                          p ,
                                                          k ,
                                                          v
                                                         ) ;
          _ctrlpts.push_back(
                             glp_get_col_prim(
                                              lp ,
                                              int( c ) + 1
                                             )
                            ) ;
        }
      }
    }
    
    /*
     * Obtain the lower and upper bounds of the second differences.
     */
    for ( unsigned p = 0 ; p < _np ; p++ ) {
      for ( unsigned k = 1 ; k < _dg ; k++ ) {
        for ( unsigned l = 0 ; l < 2 ; l++ ) {
          for ( unsigned v = 0 ; v < 2 ; v++ ) {
            unsigned c = compute_second_difference_column_index(
                                                                p ,
                                                                k ,
                                                                l ,
                                                                v
                                                               ) ;
            
            _secdiff.push_back(
                               glp_get_col_prim(
                                                lp ,
                                                int( c ) + 1
                                               )
                              ) ;
          }
        }
      }
    }

    /*
     * Obtain the minimum value of the objective function.
     */
    _ofvalue = glp_get_obj_val( lp ) ;
    
    return ;
  }

}

/** @} */ //end of group class.
