/** 
 * \file buildcurve2d.hpp
 *
 * \brief Definition  of a class  for threading  a C1 spline  curve of
 * degree d  through a planar channel  defined by a pair  of polygonal
 * chains.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Matem&aacute;tica, \n
 * mfsiqueira at mat (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date March 2016
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef __buildcurve2d__
#define __buildcurve2d__

#include "exceptionobject.hpp"           // ExceptionObject
#include "tabulatedfunction.hpp"         // TabulatedFunction
#include "coefficient.hpp"               // Coefficient
#include "bound.hpp"                     // Bound

extern "C" {
#include "glpk.h"                        // For all GLPK functions
}

#include <vector>                        // std::vector
#include <string>                        // std::string
#include <sstream>                       // std::stringstream
#include <cassert>                       // assert


/** 
 * \defgroup BC2NameSpace Namespace bc2.
 * @{
 */

/**
 * \namespace bc2
 *
 * \brief The namespace bc2 contains the definition and implementation
 * of a  set of classes  for computing a C1  spline curve of  degree d
 * passing through a  given planar channel delimited  by two polygonal
 * chains.
 *
 */

namespace bc2 {


  /**
   * \class BuildCurve2D
   *
   * \brief  This class  provides methods  for threading  a C1  spline
   * curve of degree d through a planar channel delimited by a pair of
   * polygonal chain.
   *
   * \attention This  class is based  on a particular case  (i.e., the
   *             planar  case)  of the  method  described  by Myles  &
   *             Peters in
   *
   *             A. Myles and J. Peters,
   *             Threading splines through 3d channels
   *             Computer-Aided Design, 37(2), 139-148, 2005.
   *
   */
  
  class BuildCurve2D {
  public:
    
    // -----------------------------------------------------------------
    //
    // Type definitions
    //
    // -----------------------------------------------------------------

    /**
     * \typedef sizetype
     *
     * \brief  Defines  a type  for  the  size  of  the array  of  the
     * Cartesian coordinates of  the lower and upper  envelopes of the
     * channel.
     *
     */
    typedef std::vector< double >::size_type sizetype ;


    /**
     * \typedef constsizetype
     *
     * \brief Defines a type for the size of the array of constraints.
     *
     */
    typedef std::vector< std::vector< Coefficient > >::size_type constsizetype ;


    /**
     * \typedef coeffsizetype
     *
     * \brief Defines a type for the size of the array of coefficients.
     *
     */
    typedef std::vector< Coefficient >::size_type coeffsizetype ;


  private:

    // ---------------------------------------------------------------
    //
    // Private data members
    //
    // ---------------------------------------------------------------
  
    unsigned _np ; ///< The number of polynomial pieces of the spline curve.
    unsigned _nc ; ///< The number of c-segments in each c-piece of the channel.
    unsigned _dg ; ///< The degree of the curve to be threaded into the channel.
    bool _closed ; ///< A flag to indicate whether the channel is closed.
    std::vector< double > _lxcoords ; ///< X coordinates of the lower polygonal chain of the channel.
    std::vector< double > _lycoords ; ///< Y coordinates of the lower polygonal chain of the channel.  
    std::vector< double > _uxcoords ; ///< X coordinates of the upper polygonal chain of the channel.
    std::vector< double > _uycoords ; ///< Y coordinates of the upper polygonal chain of the channel.
    TabulatedFunction* _tf ;  ///< A pointer to the lower and upper \f$a\f$ functions.
    std::vector< std::vector< Coefficient > > _coefficients ;  ///< Coefficients of the constraints of the linear program.
    std::vector< Bound > _bounds ;  ///< Type of the constraints and their bounds.
    std::vector< double > _ctrlpts ; ///< X and Y coordinates of the control points of the resulting spline.
    std::vector< double > _secdiff ; ///< Lower and upper bounds on the second difference values.
    double _ofvalue ;  ///< Optimal value (i.e., minimum) of the objective function.
    
  public:
  
    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    
    /**
     * \fn BuildCurve2D( unsigned np , unsigned nc , bool closed , double* lx , double* ly , double* ux , double* uy , TabulatedFunction* tf ) throw( ExceptionObject ) 
     *
     * \brief Creates an instance of this class.
     *
     * \param np The number of polynomial pieces of the spline curve.
     * \param  nc The  number of  c-segments  in each  c-piece of  the
     * channel.
     * \param closed A flag to indicate whether the channel is closed.
     * \param lx A  pointer to an array with the  x-coordinates of the
     * lower polygonal chain of the channel.
     * \param ly A  pointer to an array with the  y-coordinates of the
     * lower polygonal chain of the channel.
     * \param ux A  pointer to an array with the  x-coordinates of the
     * upper polygonal chain of the channel.
     * \param uy A  pointer to an array with the  y-coordinates of the
     * upper polygonal chain of the channel.
     * \param tf A pointer to the lower and upper \f$a\f$ functions.
     *
     */
    BuildCurve2D(
                 unsigned np ,
                 unsigned nc ,
                 bool closed ,
                 double* lx ,
                 double* ly ,
                 double* ux ,
                 double* uy ,
                 TabulatedFunction* tf
                )
      throw( ExceptionObject ) ;
    
    
    /**
     * \fn BuildCurve2D( const BuildCurve2D& b )
     *
     * \brief Clones an instance of this class.
     *
     * \param b A reference to another instance of this class.
     *
     */
    BuildCurve2D( const BuildCurve2D& b ) ;
  
    
    /**
     * \fn ~BuildCurve2D()
     *
     * \brief Releases the memory held by an instance of this class.
     */
    ~BuildCurve2D() ;


    /**
     * \fn bool build( int& error )
     *
     * \brief Solves the channel problem by solving a linear program.
     *
     * \param error Code returned by the LP solver whenever a solution
     * could not be  found. If a solution is found,  this parameter is
     * ignored.
     *
     * \return The logic  value true if the LP solver  is able to find
     * an  optimal solution  for the  channel problem;  otherwise, the
     * logic value false is returned.
     *
     */
    bool build( int& error ) ;


    /**
     * \fn unsigned get_number_of_curve_pieces() const
     *
     * \brief Returns the number of curves pieces of the spline curve.
     *
     * \return The number of curves pieces of the spline curve.
     *
     */
    unsigned get_number_of_curve_pieces() const
    {
      return _np ;
    }


    /**
     * \fn unsigned get_spline_degree() const
     *
     * \brief Returns the degree of the spline curve.
     *
     * \return The degree of the spline curve.
     *
     */
    unsigned get_spline_degree() const
    {
      return _dg ;
    }


    /**
     * \fn unsigned get_number_of_constraints() const throw( ExceptionObject )
     *
     * \brief Returns the number of constraints of the instance of the
     * linear program  corresponding to the channel  problem solved by
     * this class.
     *
     * \return The number of constraints of the instance of the linear
     * program  corresponding to  the channel  problem solved  by this
     * class.
     *
     */
    unsigned get_number_of_constraints() const throw( ExceptionObject )
    {
      if ( _coefficients.empty() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "No constraint has been created so far" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }
	    
      return unsigned( _coefficients.size() ) ;
    }

 
    /**
     * \fn double get_control_value( unsigned p , unsigned i , unsigned v ) const throw( ExceptionObject ) 
     *
     * \brief Returns the v-th coordinate of the i-th control point of
     * the  p-th curve  piece of  the spline  curve threaded  into the
     * channel.
     *
     * \param p The index of the p-th curve piece of the spline.
     * \param i The index of the  i-th control point of the p-th curve
     * of the spline.
     * \param  v The  v-th Cartesian  coordinate of  the i-th  control
     * point of the p-th curve of the spline.
     *
     * \return The  v-th coordinate of  the i-th control point  of the
     * p-th curve piece of the spline curve threaded into the channel.
     *
     */
    double get_control_value(
                             unsigned p ,
                             unsigned i ,
                             unsigned v
                            )
      const throw( ExceptionObject ) 
    {
      if ( p >= _np ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Index of the curve is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }
        
      if ( i > _dg ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Index of the control point is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }
        
      if ( v >= 2 ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Index of the Cartesian coordinate is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }
        
      if ( _ctrlpts.empty() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Control points have not been computed" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      sizetype index = ( 2 * ( _dg + 1 ) * p ) + ( 2 * i ) + v ;
      
      return _ctrlpts[ index ] ;
    }


    /**
     * \fn unsigned get_number_of_coefficients_in_the_ith_constraint( unsigned i ) const throw( ExceptionObject )
     *
     * \brief  Returns   the  number  of  coefficients   of  the  i-th
     * constraint of the instance  of the linear program corresponding
     * to the channel problem solved by this class.
     *
     * \param i The index of a constraint. 
     *
     * \return The  number of coefficients  of the i-th  constraint of
     * the instance of the linear program corresponding to the channel
     * problem solved by this class.
     *
     */
    unsigned get_number_of_coefficients_in_the_ith_constraint( unsigned i ) const throw( ExceptionObject )
    {
      if ( _coefficients.empty() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "No constraint has been created so far" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( constsizetype( i ) >= _coefficients.size() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Constraint index is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      return unsigned( _coefficients[ constsizetype( i ) ].size() ) ;
    }
    
      
    /**
     * \fn unsigned get_coefficient_identifier( unsigned i , unsigned j ) const throw( ExceptionObject )
     *
     * \brief Returns the number of the column that corresponds to the
     * j-th  coefficient  of the  i-th  constraint  in the  constraint
     * matrix of the  instance of the linear  program corresponding to
     * the channel problem.
     *
     * \param i The index of a constraint.
     * \param j The j-th (nonzero) coefficient of the i-th constraint.
     *
     * \return The number  of the column that corresponds  to the j-th
     * coefficient of the i-th constraint  in the constraint matrix of
     * the instance of the linear program corresponding to the channel
     * problem.
     *
     */
    unsigned get_coefficient_identifier( unsigned i , unsigned j ) const throw( ExceptionObject )
    {
      if ( _coefficients.empty() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "No constraint has been created so far" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( constsizetype( i ) >= _coefficients.size() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Constraint index is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( coeffsizetype( j ) >= _coefficients[ constsizetype( i ) ].size() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Coefficient index is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }
	   
      return _coefficients[ constsizetype( i ) ][ coeffsizetype( j ) ].get_col() ;
    }


    /**
     * \fn unsigned get_coefficient_value( unsigned i , unsigned j ) const throw( ExceptionObject )
     *
     * \brief Returns the ( i , j ) entry of the matrix of constraints
     * of  the instance  of the  linear program  corresponding to  the
     * channel problem.
     *
     * \param i The index of a constraint.
     * \param j The j-th (nonzero) coefficient of the i-th constraint.
     *
     * \return The ( i , j ) entry of the matrix of constraints of the
     * instance  of the  linear program  corresponding to  the channel
     * problem.
     *
     */
    double get_coefficient_value( unsigned i , unsigned j ) const throw( ExceptionObject )
    {
      if ( _coefficients.empty() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "No constraint has been created so far" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( constsizetype( i ) >= _coefficients.size() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Constraint index is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( coeffsizetype( j ) >= _coefficients[ constsizetype( i ) ].size() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Coefficient index is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }
	   
      return _coefficients[ constsizetype( i ) ][ coeffsizetype( j ) ].get_value() ;
    }


    /**
     * \fn double get_bound_of_ith_constraint( unsigned i ) const throw( ExceptionObject )
     *
     * \brief Returns the real value that bounds the i-th constraint.
     *
     * \param i The index of a constraint. 
     *
     * \return The real value that bounds the i-th constraint.
     *
     */
    double get_bound_of_ith_constraint( unsigned i ) const throw( ExceptionObject )
    {
      if ( _coefficients.empty() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "No constraint has been created so far" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( constsizetype( i ) >= _coefficients.size() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Constraint index is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

#ifdef DEBUGMODE
      assert( _bounds.size() == _coefficients.size() ) ;
      assert( _bounds.size() > std::vector< std::vector< Bound > >::size_type( i ) ) ;
#endif

      return _bounds[ std::vector< std::vector< Bound > >::size_type( i ) ].get_value() ;
    }


    /**
     * \fn bool is_equality( unsigned i ) const throw( ExceptionObject )
     *
     * \brief Returns  the logic value  true if  the type of  the i-th
     * constraint  is equality;  otherwise,  returns  the logic  value
     * false.
     *
     * \param i The index of a constraint. 
     *
     * \return The logic value true if the type of the i-th constraint
     * is equality; otherwise, the logic value false is returned.
     *
     */
    bool is_equality( unsigned i ) const throw( ExceptionObject )
    {
      if ( _coefficients.empty() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "No constraint has been created so far" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( constsizetype( i ) >= _coefficients.size() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Constraint index is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

#ifdef DEBUGMODE
      assert( _bounds.size() == _coefficients.size() ) ;
      assert( _bounds.size() > std::vector< std::vector< Bound > >::size_type( i ) ) ;
#endif

      return _bounds[ std::vector< std::vector< Bound > >::size_type( i ) ].get_type() == Bound::EQT ;
    }

    /**
     * \fn bool is_greater_than_or_equal_to( unsigned i ) const throw( ExceptionObject )
     *
     * \brief Returns the  logic value true if the  i-th constraint is
     * an inequality of the type  greater than or equal to; otherwise,
     * returns the logic value false.
     *
     * \param i The index of a constraint. 
     *
     * \return  The logic  value true  if  the i-th  constraint is  an
     * inequality of the type greater than or equal to; otherwise, the
     * logic value false is returned.
     *
     */
    bool is_greater_than_or_equal_to( unsigned i ) const throw( ExceptionObject )
    {
      if ( _coefficients.empty() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "No constraint has been created so far" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( constsizetype( i ) >= _coefficients.size() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Constraint index is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

#ifdef DEBUGMODE
      assert( _bounds.size() == _coefficients.size() ) ;
      assert( _bounds.size() > std::vector< std::vector< Bound > >::size_type( i ) ) ;
#endif

      return _bounds[ std::vector< std::vector< Bound > >::size_type( i ) ].get_type() == Bound::GTE ;
    }


    /**
     * \fn bool is_less_than_or_equal_to( unsigned i ) const throw( ExceptionObject )
     *
     * \brief Returns the  logic value true if the  i-th constraint is
     * an inequality  of the  type less than  or equal  to; otherwise,
     * returns the logic value false.
     *
     * \param i The index of a constraint. 
     *
     * \return  The logic  value true  if  the i-th  constraint is  an
     * inequality of  the type less  than or equal to;  otherwise, the
     * logic value false is returned.
     *
     */
    bool is_less_than_or_equal_to( unsigned i ) const throw( ExceptionObject )
    {
      if ( _coefficients.empty() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "No constraint has been created so far" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( constsizetype( i ) >= _coefficients.size() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Constraint index is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

#ifdef DEBUGMODE
      assert( _bounds.size() == _coefficients.size() ) ;
      assert( _bounds.size() > std::vector< std::vector< Bound > >::size_type( i ) ) ;
#endif

      return _bounds[ std::vector< std::vector< Bound > >::size_type( i ) ].get_type() == Bound::LTE ;
    }

    
    /**
     * \fn double get_lower_bound_on_second_difference_value( unsigned p , unsigned i , unsigned v ) const throw( ExceptionObject ) 
     *
     * \brief Returns the lower bound (found  by the LP solver) on the
     * v-th coordinate  of the  i-th second  difference vector  of the
     * p-th curve piece of the spline curve threaded into the channel.
     *
     * \param p The index of the p-th curve piece of the spline.

     * \param i  The index of the  i-th second difference of  the p-th
     * curve of the spline.

     * \param  v The  v-th Cartesian  coordinate of  the i-th  control
     * point of the p-th curve of the spline.
     *
     * \return the  lower bound (found by  the LP solver) on  the v-th
     * coordinate of  the i-th  second difference  vector of  the p-th
     * curve piece of the spline curve threaded into the channel.
     *
     */
    double get_lower_bound_on_second_difference_value(
                                                      unsigned p ,
                                                      unsigned i ,
                                                      unsigned v
                                                     )
      const throw( ExceptionObject ) 
    {
      if ( p >= _np ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Index of the curve is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( ( i < 1 ) || ( i > _dg ) ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Index of the second difference vector is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( v >= 2 ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Index of the Cartesian coordinate is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( _secdiff.empty() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Second differences have not been computed" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      sizetype index = ( 4 * ( _dg - 1 ) * p ) + ( 4 * ( i - 1 ) ) + v ;
      
      return _secdiff[ index ] ;
    }


    /**
     * \fn double get_upper_bound_on_second_difference_value( unsigned p , unsigned i , unsigned v ) const throw( ExceptionObject ) 
     *
     * \brief Returns the upper bound (found  by the LP solver) on the
     * v-th coordinate  of the  i-th second  difference vector  of the
     * p-th curve piece of the spline curve threaded into the channel.
     *
     * \param p The index of the p-th curve piece of the spline.

     * \param i  The index of the  i-th second difference of  the p-th
     * curve of the spline.

     * \param  v The  v-th Cartesian  coordinate of  the i-th  control
     * point of the p-th curve of the spline.
     *
     * \return The  upper bound (found by  the LP solver) on  the v-th
     * coordinate of  the i-th  second difference  vector of  the p-th
     * curve piece of the spline curve threaded into the channel.
     *
     */
    double get_upper_bound_on_second_difference_value(
                                                      unsigned p ,
                                                      unsigned i ,
                                                      unsigned v
                                                     )
      const throw( ExceptionObject )
    {
      if ( p >= _np ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Index of the curve is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( ( i < 1 ) || ( i > _dg ) ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Index of the second difference vector is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      if ( v >= 2 ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Index of the Cartesian coordinate is out of range" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }
      
      if ( _secdiff.empty() ) {
        std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
        ss << "Second differences have not been computed" ;
        throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
      }

      sizetype index = ( 4 * ( _dg - 1 ) * p ) + ( 4 * ( i - 1 ) ) + 2 + v ;
      
      return _secdiff[ index ] ;
    }


    /**
     * \fn double minimum_value() const
     *
     * \brief  Returns the  optimal (minimum)  value of  the objective
     * function of the instance of the channel problem as found by the
     * LP solver.
     *
     * \return The  optimal (minimum) value of  the objective function
     * of  the instance  of the  channel problem  as found  by the  LP
     * solver.
     *
     */
    double minimum_value() const
    {
      return _ofvalue ;
    }
        
        
    /**
     * \fn std::string get_solver_error_message( int error )
     *
     * \brief Returns the error message of the GLPK solver associated with a given error code.
     *
     * \param error Error code returned by the LP solver.
     *
     * \return The error message of the GLPK solver associated with a given error code.
     *
     */
     std::string get_solver_error_message( int error )
     {
       std::string message ;
       switch ( error ) {
         case GLP_EBADB :
           message = "Unable to start the search because the number of basic variables is not the same as the number of rows in the problem object." ;
           break ;
         case GLP_ESING :
           message = "Unable to start the search because the basis matrix corresponding to the initial basis is singular within the working precision." ;
           break ;
         case GLP_ECOND :
           message = "Unable to start the search because the basis matrix corresponding to the initial basis is ill-conditioned." ;
           break ;
         case GLP_EBOUND :
           message = "Unable to start the search because some double-bounded variables have incorrect bounds." ;
           break ;
         case GLP_EFAIL :
           message = "The search was prematurely terminated due to the solver failure." ;
           break ;
         case GLP_EOBJLL :
           message = "The search was prematurely terminated because the objective function being maximized has reached its lower limit and continues decreasing." ;
           break ;
         case GLP_EOBJUL :
           message = "The search was prematurely terminated because the objective function being minimized has reached its upper limit and continues increasing." ;
           break ;
         case GLP_EITLIM :
           message = "The search was prematurely terminated because the simplex iteration limit has been exceeded." ;
           break ;
         case GLP_ETMLIM :
           message = "The search was prematurely terminated because the time limit has been exceeded." ;
           break ;
         case GLP_ENOPFS :
           message = "The LP problem instance has no primal feasible solution." ;
           break ;
         case GLP_ENODFS :
           message = "The LP problem instance has no dual feasible solution." ;
           break ;
         default :
           message = "Unknown reason." ;
           break ;
       }
       
       return message ;
     }

  
  private:
  
    // ---------------------------------------------------------------
    //
    // Private methods
    //
    // ---------------------------------------------------------------


    /**
     * \fn double BuildCurve2D::lf( double u , double b0 , double bd ) const
     *
     * \brief Computes the value of  the affine function \f$\ell\f$ at
     * a given point of the interval \f$[0,1]\f$ of the real line.
     *
     * \param u A parameter point in the interval \f$[0,1]\f$.
     * \param b0 The x- or y-coordinate  of the first control point of
     * a B&eacute;zier curve of degree d.
     * \param bd The x- or y-coordinate of the last control point of a
     * B&eacute;zier curve of degree d.
     *
     * \return The value of the  affine function \f$\ell\f$ at a given
     * point of the interval \f$[0,1]\f$ of the real line.
     */
    double lf(
              double u ,
              double b0 ,
              double bd
             )
    const ;


    /**
     * \fn double h( double u ) const
     *
     * \brief Computes the value of a piecewise affine hat function at
     * a given point of the real line.
     *
     * \param u A parameter point in the real line.
     *
     * \return The value of a piecewise linear hat function at a given
     * point of the real line.
     *
     */
    double h( double u ) const ;


    /**
     * \fn void compute_normal_to_lower_envelope( unsigned s , double& nx , double& ny ) const
     *
     * \brief Computes an  outward normal to the s-th  line segment of
     * the lower envelope of the channel.
     *
     * \param s Index of a line segment of the lower channel envelope.
     * \param nx A reference to  the first Cartesian coordinate of the
     * normal.
     * \param ny A reference to the Second Cartesian coordinate of the
     * normal.
     *
     */
    void compute_normal_to_lower_envelope(
                                          sizetype s ,
                                          double& nx ,
                                          double& ny
                                         )
      const ;


    /**
     * \fn void compute_normal_to_upper_envelope( unsigned s , double& nx , double& ny ) const
     *
     * \brief Computes an  outward normal to the s-th  line segment of
     * the upper envelope of the channel.
     *
     * \param s Index of a line segment of the upper channel envelope.
     * \param nx A reference to  the first Cartesian coordinate of the
     * normal.
     * \param ny A reference to the Second Cartesian coordinate of the
     * normal.
     *
     */
    void compute_normal_to_upper_envelope(
                                          sizetype s ,
                                          double& nx ,
                                          double& ny
                                         )
      const ;


    /**
     * \fn unsigned compute_control_value_column_index( unsigned p , unsigned i , unsigned v ) const
     *
     * \brief Computes the  index of the linear  program matrix column
     * corresponding to  the x-  or y-coordinate  of the  i-th control
     * point of the p-th piece of the spline curve to be threaded into
     * the channel.
     *
     * \param p Index of the spline curve piece.
     * \param i  Index of  a control  point of  the p-th  spline curve
     * piece.
     * \param v Index of the x- or y-coordinate of the control point.
     *
     * \return  The   index  of  the  linear   program  matrix  column
     * corresponding to  the x-  or y-coordinate  of the  i-th control
     * point of the p-th piece of the spline curve to be threaded into
     * the channel.
     *
     */
    unsigned compute_control_value_column_index(
                                                unsigned p ,
                                                unsigned i ,
                                                unsigned v 
                                               )
      const ;


    /**
     * \fn unsigned compute_second_difference_column_index( unsigned p , unsigned i , unsigned l , unsigned v ) const
     *
     * \brief Computes the  index of the linear  program matrix column
     * corresponding to  the x- or  y-coordinate of the l-th  bound of
     * the  i-th second  difference of  the p-th  piece of  the spline
     * curve to be threaded into the channel.
     *
     * \param p Index of the spline curve piece.
     * \param  i Index  of the  second difference  of the  p-th spline
     * curve piece.
     * \param l Index of the l-th  bound of the second difference (0 -
     * lower bound; 1 - upper bound).
     * \param  v  Index  of  the  x- or  y-coordinate  of  the  second
     * difference bound.
     *
     * \return  The   index  of  the  linear   program  matrix  column
     * corresponding to  the x- or  y-coordinate of the l-th  bound of
     * the  i-th second  difference of  the p-th  piece of  the spline
     * curve to be threaded into the channel.
     *
     */
    unsigned compute_second_difference_column_index(
                                                    unsigned p ,
                                                    unsigned i ,
                                                    unsigned l ,
                                                    unsigned v 
                                                   )
      const ;
    
    
    /**
     * \fn void compute_min_max_constraints( unsigned& eqline )
     *
     * \brief Computes the equations defining the min-max constraints.
     *
     * \param eqline A reference to the counter of equations.
     *
     */
    void compute_min_max_constraints( unsigned& eqline ) ;


    /**
     * \fn void compute_correspondence_constraints( unsigned& eqline )
     *
     * \brief  Computes  the  equations  defining  the  correspondence
     * constraints.
     *
     * \param eqline A reference to the counter of equations. 
     *
     */
    void compute_correspondence_constraints( unsigned& eqline ) ;


    /**
     * \fn void compute_c0continuity_constraints( unsigned& eqline )
     *
     * \brief  Computes  the   equations  defining  the  C0-continuity
     * constraints.
     *
     * \param eqline A reference to the counter of equations. 
     *
     */
    void compute_c0continuity_constraints( unsigned& eqline ) ;


    /**
     * \fn void compute_c1continuity_constraints( unsigned& eqline )
     *
     * \brief  Computes  the   equations  defining  the  C1-continuity
     * constraints.
     *
     * \param eqline A reference to the counter of equations. 
     *
     */
    void compute_c1continuity_constraints( unsigned& eqline ) ;


    /**
     * \fn void compute_sleeve_corners_in_channel_constraints( unsigned& eqline )
     *
     * \brief  Computes the  equations defining  the constraints  that
     * ensure that the corners of the sleeves are inside the channel.
     *
     * \param eqline A reference to the counter of equations. 
     *
     */
    void compute_sleeve_corners_in_channel_constraints( unsigned& eqline ) ;


    /**
     * \fn void compute_channel_corners_outside_sleeve_constraints( unsigned& eqline )
     *
     * \brief  Computes the  equations defining  the constraints  that
     * ensure that the corners of  the channel are located outside the
     * sleeve.
     *
     * \param eqline A reference to the counter of equations. 
     *
     */
    void compute_channel_corners_outside_sleeve_constraints( unsigned& eqline ) ;


    /**
     * \fn int solve_lp( sizetype rows , sizetype cols )
     *
     * \brief Solves  the linear program corresponding  to the channel
     * problem.
     *
     * \param rows The number of constraints of the linear program.
     * \param cols The number of unknowns of the linear program.
     *
     * \return  The code  returned by  the LP  solver to  indicate the
     * status  of  the  computation  of the  solution  of  the  linear
     * program.
     */
    int solve_lp(
                  sizetype rows ,
                  sizetype cols
                 )  ;


    /**
     * \fn void set_up_lp_constraints( glp_prob* lp ) const
     *
     * \brief Assemble the matrix of  constraints of the linear program,
     * and define  the type (equality  or inequality) and bounds  on the
     * constraints.
     *
     * \param lp A pointer to the instance of the LP program.
     *
     */
    void set_up_lp_constraints( glp_prob* lp ) const ;


    /**
     * \fn void BuildCurve2D::set_up_structural_variables( glp_prob* lp ) const
     *
     * \brief  Define  lower and/or  upper  bounds  on the  structural
     * variables of  the linear  program corresponding to  the channel
     * problem.
     *
     * \param lp A pointer to the instance of the LP program.
     *
     */
    void set_up_structural_variables( glp_prob* lp ) const ;


    /**
     * \fn void BuildCurve2D::set_up_objective_function( glp_prob* lp ) const
     *
     * \brief  Define the  objective  function of  the linear  program
     * corresponding to  the channel problem, which  is a minimization
     * problem.
     *
     * \param lp A pointer to the instance of the LP program.
     *
     */
    void set_up_objective_function( glp_prob* lp ) const ;


    /**
     * \fn void BuildCurve2D::get_lp_solver_result_information( glp_prob* lp )
     *
     * \brief Obtain the optimal values found by the LP solver for the
     * structural values  of the  linear programming  corresponding to
     * the channel problem.
     *
     * \param lp A pointer to the instance of the LP program.
     *
     */
    void get_lp_solver_result_information( glp_prob* lp ) ;
    

  } ;
  

}

/** @} */ //end of group class.

#endif   // __buildcurve2d__
