/** 
 * \file coefficient.hpp
 *
 * \brief Definition of a class for representing a nonzero coefficient
 * of a variable of a linear constraint (inequality or equality) of an
 * LP.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Matem&aacute;tica, \n
 * mfsiqueira at mat (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date March 2016
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef __coefficient__
#define __coefficient__


/** 
 * \defgroup BC2NameSpace Namespace bc2.
 * @{
 */

/**
 * \namespace bc2
 *
 * \brief The namespace bc2 contains the definition and implementation
 * of a  set of  classes for  computing a C1  spline cubic  curve that
 * passes through  a given planar  channel delimited by  two polygonal
 * chains.
 *
 */

namespace bc2 {

  /**
   * \class Coefficient
   *
   * \brief This class represents a  nonzero coefficient of a variable
   * of  a linear  constraint  (inequality or  equality)  of a  linear
   * program.
   *
   */
  
  class Coefficient {
  protected:

    // ---------------------------------------------------------------
    //
    // Private data
    //
    // ---------------------------------------------------------------

    unsigned _row ;  ///< The identifier of the constraint this coefficient belongs to.
    unsigned _col ;  ///< The identifier of the unknown multiplied by this coefficient.
    double _value ;  ///< The coefficient value.


  public:
  
    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------
    
    /**
     * \fn Coefficient()
     *
     * \brief Creates an instance of this class.
     *
     */
    Coefficient()
    :
      _row( 0 ) ,
      _col( 0 ) ,
      _value( 0 )
    {
    }

    
    /**
     * \fn Coefficient( unsigned row , unsigned col , double value )
     *
     * \brief Creates an instance of this class.
     *
     * \param row The identifier of the  constraint  this  coefficient
     * belongs to.
     * \param col  The identifier  of the  unknown multiplied  by this
     * coefficient.
     * \param value The value of the coefficient.
     *
     */
    Coefficient( unsigned row , unsigned col , double value )
    :
      _row( row ) ,
      _col( col ) ,
      _value( value )
    {
    }


    /**
     * \fn Coefficient( const Coefficient& c )
     *
     * \brief Creates an instance of  this class from another instance
     * of this class.
     *
     * \param c An instance of this class.
     *
     */
    Coefficient( const Coefficient& c )
    :
      _row( c._row ) ,
      _col( c._col ) ,
      _value( c._value )
    {
    }
     
    
    /**
     * \fn ~Coefficient( )
     *
     * \brief Releases the memory held by an instance of this class.
     *
     */
    ~Coefficient( )
    {
    }

    
    /**
     * \fn unsigned get_row() const 
     *
     * \brief Returns the identifier of the constraint the coefficient
     * is associated  with. This identifier corresponds  to the number
     * of  a row  in the  constraint  coefficient matrix  of a  linear
     * program.
     *
     *
     * \return  The identifier  of the  constraint the  coefficient is
     * associated with.
     *
     */
    unsigned get_row() const
    {
      return _row ;
    }


    /**
     * \fn unsigned get_col() const 
     *
     * \brief Returns the identifier of the unknown multiplied by this
     * coefficient  in  a  constraint   of  a  linear  program.   This
     * identifier  corresponds  to  the  number of  a  column  in  the
     * constraint coefficient matrix of the linear program.
     *
     *
     * \return  The  identifier  of  the unknown  multiplied  by  this
     * coefficient in a constraint of a linear program.
     *
     */
    unsigned get_col() const
    {
      return _col ;
    }


    /**
     * \fn double get_value() const 
     *
     * \brief Returns the value of this coefficient.
     *
     * \return The value of this coefficient.
     *
     */
    double get_value() const
    {
      return _value ;
    }
    
  } ;

}

/** @} */ //end of group class.

#endif   // __coefficient__
