/** 
 * \file main.cpp
 *
 * \brief A simple program for testing the bc2d library.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Matem&aacute;tica, \n
 * mfsiqueira at mat (dot) ufrn (dot) br
 *
 * \version 1.0
 * \date March 2016
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include <iostream>                 // std::cout, std::endl, std::cerr
#include <fstream>                  // std::ifstream, std::ofstream
#include <string>                   // std::string
#include <cstdlib>                  // exit, EXIT_SUCCESS, EXIT_FAILURE
#include <iomanip>                  // std::setprecision
#include <cassert>                  // assert
#include <ctime>                    // time, clock, CLOCKS_PER_SEC, clock_t
#include <cstddef>                  // size_t
#include <cmath>                    // fabs

#include "exceptionobject.hpp"      // bc2::ExceptionObject
#include "tabulatedfunction.hpp"    // bc2::TabulatedFunction
#include "a3.hpp"                   // bc2::a3
#include "buildcurve2d.hpp"         // bc2::BuildCurve2D


using std::cin  ;
using std::cout ;
using std::cerr ;
using std::endl ;
using std::string ;
using bc2::TabulatedFunction ;
using bc2::a3 ;
using bc2::BuildCurve2D ;
using bc2::ExceptionObject ;



/**
 * \fn void read_input( const string& fn , unsigned& np , unsigned& nc , bool& closed , unsigned& dg , double*& lx , double*& ly , double*& ux , double*& uy ) throw( ExceptionObject ) 
 *
 * \brief Read in a file describing a polygonal channel.
 *
 * \param fn The name of a file describing a polygonal channel.
 * \param np A reference  to the number of pieces of  the spline to be
 * threaded into the channel.
 * \param nc A  reference to the number of c-segments  of each c-piece
 * of the channel.
 * \param closed A reference to a flag to indicate whether the channel
 * is closed.
 * \param  dg Degree  of  the spline  curve to  be  threaded into  the
 * channel.
 * \param  lx  A  reference  to  a   pointer  to  an  array  with  the
 * x-coordinates of the lower polygonal chain of the channel.
 * \param  ly  A  reference  to  a   pointer  to  an  array  with  the
 * y-coordinates of the lower polygonal chain of the channel.
 * \param  ux  A  reference  to  a   pointer  to  an  array  with  the
 * x-coordinates of the upper polygonal chain of the channel.
 * \param  uy  A  reference  to  a   pointer  to  an  array  with  the
 * y-coordinates of the upper polygonal chain of the channel.
 *
 */
void read_input(
                const string& fn ,
                unsigned& np ,
                unsigned& nc ,
                bool& closed ,
                unsigned& dg ,
                double*& lx ,
                double*& ly ,
                double*& ux ,
                double*& uy
               )
  throw( ExceptionObject ) ;


/**
 * \fn void write_solution( const string& fn , const BuildCurve2D& b )
 *
 * \brief Write the control points of the spline curve threaded into a
 * channel to an output file.
 *
 * \param fn The name of the output file.
 * \param b An instance of the spline curve builder.
 *
 */
void write_solution(
                    const string& fn ,
                    const BuildCurve2D& b
                   ) ;


/**
 * \fn void write_lp( const string& fn , const BuildCurve2D& b )
 *
 * \brief Write the  instance of the linear program  problem solved by
 * this program in  CPLEX format. The output file can  be given to the
 * \e gpsolve function of the GNU GLPK or to debug the assembly of the
 * constraints.
 *
 * \param fn The name of the output file.
 * \param b An instance of the spline curve builder.
 *
 */
void write_lp(
              const string& fn ,
              const BuildCurve2D& b
             ) ;



/**
 * \fn int main( int argc , char* argv[]  )
 *
 * \brief A simple program for testing the bc2d library.
 *
 * \param argc The number of command-line arguments.
 * \param argv An array with the command-line arguments.
 *
 * \return An integer number.
 */
int main( int argc , char* argv[] ) {
  //
  // Check command-line arguments.
  //
  
  if ( ( argc != 3 ) && ( argc != 4 ) ) {
    cerr << "Usage: "
         << endl
         << "\t\t test-bc2d arg1 arg2 [ arg3 ]"
         << endl
         << "\t\t arg1: name of the file describing the polygonal channel"
         << endl
         << "\t\t arg2: name of the output file describing the computed spline curve"
         << endl
         << "\t\t arg3: name of an output file to store a CPLEX format definition of the LP solved by this program (OPTIONAL)"
         << endl
         << endl ;
    cerr.flush() ;
    
    return EXIT_FAILURE ;
  }
  
  //
  // Read in the input file.
  //
  
  clock_t start, end ;
  
  cerr << endl
       << "Reading file describing a polygonal channel..."
       << endl ;
  cerr.flush() ;
  
  string fn1( argv[ 1 ] ) ;

  unsigned np ;
  unsigned nc ;
  bool closed ;
  unsigned dg ;
  double* lx ;
  double* ly ;
  double* ux ;
  double* uy ;

  start = clock() ;
  try {
    read_input( fn1 , np , nc , closed , dg , lx , ly , ux , uy ) ;
  }
  catch ( const ExceptionObject& xpt ) {
    treat_exception( xpt ) ;
    exit( EXIT_FAILURE ) ;
  }  
  end = clock() ;
  
  cerr << ( (double) ( end - start ) ) / CLOCKS_PER_SEC
       << " seconds."
       << endl
       << endl ;
  cerr.flush() ;
  
  //
  // Compute a cubic spline curve that passes through the channel.
  //
  
  cerr << "Computing a cubic spline curve that passes through the channel... "
       << endl ;
  cerr.flush() ;
  
  start = clock() ;
  assert( dg == 3 ) ;
  TabulatedFunction* tf = new a3() ;
  BuildCurve2D* builder = 0 ;
  try {
    builder = new BuildCurve2D(
                               np ,
                               nc ,
                               closed ,
                               &lx[ 0 ] ,
                               &ly[ 0 ] ,
                               &ux[ 0 ] ,
                               &uy[ 0 ] ,
                               tf
                              ) ;
  }
  catch ( const ExceptionObject& xpt ) {
    treat_exception( xpt ) ;
    exit( EXIT_FAILURE ) ;
  }  
  int error ;
  bool res = builder->build( error ) ;
  end = clock() ;
  
  cerr << ( (double) ( end - start ) ) / CLOCKS_PER_SEC
       << " seconds."
       << endl
       << endl ;
  cerr.flush() ;

  if ( res ) {
    //
    // Write the control points of all spline pieces to a file.
    //
    
    cerr << "Writing out the control points of all spline pieces to a file..."
	       << endl ;
    cerr.flush() ;
  
    start = clock() ;  
    string fn2( argv[ 2 ] ) ;
    write_solution(
                   fn2 ,
                   *builder
                  ) ;
    end = clock() ;
  
    cerr << ( (double) ( end - start ) ) / CLOCKS_PER_SEC
         << " seconds."
         << endl
         << endl ;
    cerr.flush() ;
  }
  else {
    //
    // Print the error message returned by the LP solver.
    //
    cerr << endl
         << "ATTENTION: "
	 << endl
         << builder->get_solver_error_message( error )
         << endl
         << endl ;
  }

  //
  // Generate a description of the linear program in CPLEX format.
  //
  if ( argc == 4 ) {
    cerr << "Writing out a description of the linear program in CPLEX format..."
	       << endl ;
    cerr.flush() ;
  
    start = clock() ;  
    string fn3( argv[ 3 ] ) ;
    write_lp(
             fn3 ,
             *builder
            ) ;
    end = clock() ;
  
    cerr << ( (double) ( end - start ) ) / CLOCKS_PER_SEC
         << " seconds."
         << endl
         << endl ;
    cerr.flush() ;
  }
  
  //
  // Release memory
  //
  
  cerr << "Releasing memory..."
       << endl ;
  cerr.flush() ;
  
  start = clock() ;
  if ( lx != 0 ) delete[ ] lx ;
  if ( ly != 0 ) delete[ ] ly ;
  if ( ux != 0 ) delete[ ] ux ;
  if ( uy != 0 ) delete[ ] uy ;
  if ( builder != 0 ) delete builder ;
  end = clock() ;
  
  cerr << ( (double) ( end - start ) ) / CLOCKS_PER_SEC
  << " seconds."
  << endl
  << endl ;
  cerr.flush() ;
  
  //
  // Done.
  //
  
  cerr << "Finished."
       << endl
       << endl
       << endl ;
  cerr.flush() ;
  
  return EXIT_SUCCESS ;
}


/**
 * \fn void read_input( const string& fn , unsigned& np , unsigned& nc , bool& closed , unsigned& dg , double*& lx , double*& ly , double*& ux , double*& uy ) throw( ExceptionObject ) 
 *
 * \brief Read in a file describing a polygonal channel.
 *
 * \param fn The name of a file describing a polygonal channel.
 * \param np A reference  to the number of pieces of  the spline to be
 * threaded into the channel.
 * \param nc A  reference to the number of c-segments  of each c-piece
 * of the channel.
 * \param closed A reference to a flag to indicate whether the channel
 * is closed.
 * \param  dg Degree  of  the spline  curve to  be  threaded into  the
 * channel.
 * \param  lx  A  reference  to  a   pointer  to  an  array  with  the
 * x-coordinates of the lower polygonal chain of the channel.
 * \param  ly  A  reference  to  a   pointer  to  an  array  with  the
 * y-coordinates of the lower polygonal chain of the channel.
 * \param  ux  A  reference  to  a   pointer  to  an  array  with  the
 * x-coordinates of the upper polygonal chain of the channel.
 * \param  uy  A  reference  to  a   pointer  to  an  array  with  the
 * y-coordinates of the upper polygonal chain of the channel.
 *
 */
void read_input(
                const string& fn ,
                unsigned& np ,
                unsigned& nc ,
                bool& closed ,
                unsigned& dg ,
                double*& lx ,
                double*& ly ,
                double*& ux ,
                double*& uy
               )
  throw( ExceptionObject ) 
{
  //
  // Open the input file
  //
  std::ifstream in( fn.c_str() ) ;
  
  if ( in.is_open() ) {
    //
    // Read in the number of polynomial pieces.
    //
    in >> np ;

    //
    // Read in the number of c-segments of each c-piece of the channel.
    //
    in >> nc ;

    //
    // Read in the flag indicating whether the channel is closed.
    //
    unsigned flag ;
    in >> flag ;

    if ( ( flag != 0 ) && ( flag != 1 ) ) {
      std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
      ss << "Flag value indicating whether the channel is closed or open is invalid" ;
      in.close() ;
      throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
    }

    closed = ( flag == 1 ) ;
    
    //
    // Read in the number of points in each polygonal chain.
    //
    unsigned nn ;
    in >> nn ;

      if (
          (  closed && ( nn != (   np * nc )       ) )
          ||
          ( !closed && ( nn != ( ( np * nc ) + 1 ) ) )
         )
    {
      std::stringstream ss( std::stringstream::in | std::stringstream::out ) ;
      ss << "Inconsistency among the number of channel breakpoints, curve pieces, and c-segments per c-piece" ;
      in.close() ;
      throw ExceptionObject( __FILE__ , __LINE__ , ss.str().c_str() ) ;
    }
    
    //
    // Read in the degree of the spline to be threaded into the channel.
    //
    in >> dg ;
    
    //
    // Read in  the coordinates of  the points of the  lower polygonal
    // chain.
    //
    
    lx = new double[ nn ] ;
    ly = new double[ nn ] ;

    for ( unsigned i = 0 ; i < nn ; i++ ) {
      //
      // Read in the X and Y coordinates of the i-th point.
      //
      in >> lx[ i ] ;
      in >> ly[ i ] ;
    }

    //
    // Read in  the coordinates of  the points of the  upper polygonal
    // chain.
    //
    
    ux = new double[ nn ] ;
    uy = new double[ nn ] ;

    for ( unsigned i = 0 ; i < nn ; i++ ) {
      //
      // Read in the X and Y coordinates of the i-th point.
      //
      in >> ux[ i ] ;
      in >> uy[ i ] ;
    }
    
    //
    // Close file
    //
    
    in.close() ;
  }
  
  return ;
}


/**
 * \fn void write_solution( const string& fn , const BuildCurve2D& b )
 *
 * \brief Write the control points of the spline curve threaded into a
 * channel to an output file.
 *
 * \param fn The name of the output file.
 * \param b An instance of the spline curve builder.
 *
 */
void write_solution(
                    const string& fn ,
                    const BuildCurve2D& b
                   )
{
  using std::endl ;
  
  std::ofstream ou( fn.c_str() ) ;
  
  if ( ou.is_open() ) {
    //
    // Set the precision of the floating-point numbers.
    //
    
    ou << std::setprecision( 6 ) << std::fixed ;
    
    //
    // Write the number of curve pieces and the degree of the spline.
    //
    
    unsigned np = b.get_number_of_curve_pieces() ;
    unsigned dg = b.get_spline_degree() ;
    
    ou << np
       << '\t'
       << dg
       << endl ;
    
    for ( unsigned p = 0 ; p < np ; p++ ) {
      for ( unsigned i = 0 ; i <= dg ; i++ ) {
        double x ;
        double y ;
        try {
          x = b.get_control_value( p , i , 0 ) ;
          y = b.get_control_value( p , i , 1 ) ;
        }
        catch ( const ExceptionObject& xpt ) {
          treat_exception( xpt ) ;
          ou.close() ;
          exit( EXIT_FAILURE ) ;
        }
        ou << x
           << '\t'
           << y
           << endl ;
      }
    }
    
    //
    // Close file
    //
    
    ou.close() ;
  }

  return ;
}


/**
 * \fn void write_lp( const string& fn , const BuildCurve2D& b )
 *
 * \brief Write the  instance of the linear program  problem solved by
 * this program in  CPLEX format. The output file can  be given to the
 * \e gpsolve function of the GNU GLPK or to debug the assembly of the
 * constraints.
 *
 * \param fn The name of the output file.
 * \param b An instance of the spline curve builder.
 *
 */
void write_lp(
              const string& fn ,
              const BuildCurve2D& b
             )
{
  //
  // Create the output file
  //

  const unsigned np = b.get_number_of_curve_pieces() ;
  const unsigned dg = b.get_spline_degree() ;

  const unsigned NUMCPC = 2 * ( dg + 1 ) ;
  const unsigned NUMSDC = 4 * ( dg - 1 ) ;
  const unsigned NUMCRTLPTS = np * NUMCPC ;
  const unsigned NUMSECDIFF = np * NUMSDC ;

  std::ofstream ou( fn.c_str() ) ;
  
  if ( ou.is_open() ) {
    //
    // Set the precision of the floating-point numbers.
    //
    
    ou << std::setprecision( 6 ) << std::fixed ;

    //
    // Write the objective function
    //
    
    ou << "Minimize"
       << std::endl ;
    ou << '\t'
       << "obj: " ;
    
    unsigned j = 1 ;
    for ( unsigned i = 0 ; i < NUMSECDIFF ; i += 4 ) {
      ou << "-mx" << j << " " ;
      ou << "-my" << j << " " ;
      ou << "+px" << j << " " ;
      ou << "+py" << j << " " ;
      ++j ;
    }

    ou << std::endl ;
    
    //
    // Write the constraints
    //

    ou << "Subject To" << std::endl ;
    
    try {
      
      for( unsigned i = 0 ; i < b.get_number_of_constraints() ; i++ ) {
        /*
         * Write out the number of the constraint.
         */
        ou << '\t' << "c" << i + 1 << ": " ;
        
        /*
         * Get the coefficients of the i-th constraint.
         */
        for( j = 0 ; j < b.get_number_of_coefficients_in_the_ith_constraint( i ) ; ++j ) {
          /*
           * Get the column index of the coefficient.
           */
          unsigned col = b.get_coefficient_identifier( i , j ) ;
          
          /*
           * Get the value of the coefficient.
           */
          double value = b.get_coefficient_value( i , j ) ;
          
          /*
           * Compute the  index of the  curve piece associated  with the
           * coefficient, and  find the type of  structural variable the
           * coefficient is.
           */
          unsigned pth = col / ( NUMCPC + NUMSDC ) ;
          unsigned var = col % ( NUMCPC + NUMSDC ) ;
          if ( var < NUMCPC ) {
            unsigned co = pth * ( dg + 1 ) + ( var / 2 ) + 1 ;
            if ( ( var % 2 ) == 0 ) {
              if ( value > 0 ) {
                ou << "+" << value << "x" << co << " " ;
              }
              else {
                ou << value << "x" << co << " " ;
              }
            }
            else {
              if ( value > 0 ) {
                ou << "+" << value << "y" << co << " " ;
              }
              else {
                ou << value << "y" << co << " " ;
              }
            }
          }
          else {
            var -= NUMCPC ;
            unsigned co = pth * ( dg - 1 ) + ( var / 4 ) + 1 ;
            unsigned ro = var % 4 ;
            if ( ro == 0 ) {
              if ( value > 0 ) {
                ou << "+" << value << "mx" << co << " " ;
              }
              else {
                ou << value << "mx" << co << " " ;
              }
            }
            else if ( ro == 1 ) {
              if ( value > 0 ) {
                ou << "+" << value << "my" << co << " " ;
              }
              else {
                ou << value << "my" << co << " " ;
              }
            }
            else if ( ro == 2 ) {
              if ( value > 0 ) {
                ou << "+" << value << "px" << co << " " ;
              }
              else {
                ou << value << "px" << co << " " ;
              }
            }
            else if ( ro == 3 ) {
              if ( value > 0 ) {
                ou << "+" << value << "py" << co << " " ;
              }
              else {
                ou << value << "py" << co << " " ;
              }
            }
          }
        }
        
        if ( b.is_equality( i ) ) {
          ou << " = " ;
        }
        else if ( b.is_less_than_or_equal_to( i ) ) {
          ou << " <= " ;
        }
        else {
#ifdef DEBUGMODE
          assert( b.is_greater_than_or_equal_to( i ) ) ;
#endif
          ou << " >= " ;
        }
        
        ou << b.get_bound_of_ith_constraint( i ) << std::endl ;
      }
      
    }
    catch ( const ExceptionObject& xpt ) {
      treat_exception( xpt ) ;
      exit( EXIT_FAILURE ) ;
    }
  
    //
    // Write the bounds
    //
    
    ou << "Bounds" << std::endl ;
    
    for ( unsigned k = 0 ; k < NUMCRTLPTS ; k += 2 ) {
      unsigned h = ( k >> 1 ) + 1 ;
      ou << '\t' << "x" << h << " free" << std::endl ;
      ou << '\t' << "y" << h << " free" << std::endl ;
    }
    
    for ( unsigned k = 0 ; k < NUMSECDIFF ; k += 4 ) {
      unsigned h = ( k >> 2 ) + 1 ;
      ou << '\t' << "-inf <= mx" << h << " <= 0" << std::endl ;
      ou << '\t' << "-inf <= my" << h << " <= 0" << std::endl ;
      ou << '\t' << "0 <= px" << h << " <= +inf" << std::endl ;
      ou << '\t' << "0 <= py" << h << " <= +inf" << std::endl ;
    }
    
    ou << "End" << std::endl ;
    
    //
    // Close file
    //
    
    ou.close() ;
  }
  
  return ;
}
